import torch
from torch import optim, nn
import torch.nn.functional as F
import lightning.pytorch as pl
from bento_sc.utils.config import Config
from bento_sc.utils.metrics import pearson_batch_masked
from bento_sc import loss
from scipy.stats import pearsonr
import numpy as np
from torchmetrics.classification import MulticlassAccuracy
from copy import deepcopy


class Permute(nn.Module):
    def __init__(self, *args):
        super().__init__()
        self.args = args

    def forward(self, x):
        return x.permute(*self.args)


class View(nn.Module):
    def __init__(self, *args):
        super().__init__()
        self.args = args

    def forward(self, x):
        return x.reshape(*self.args)


class PerturbBaseline(pl.LightningModule):
    def __init__(
        self,
        config,
    ):
        super().__init__()
        self.save_hyperparameters()

        self.config = deepcopy(config)
        dim_per_gene = self.config.baseline_perturb_dim_per_gene
        bottleneck_dim = self.config.baseline_perturb_bottleneck_dim

        self.perturbation_indicator = nn.Parameter(
            torch.empty(dim_per_gene).uniform_(-1, 1) / self.config.perturb_init_factor
        )

        self.to_embed = nn.Linear(1, dim_per_gene)  # B, G, 1 -> B, G, H1
        self.mixer = nn.Sequential(
            nn.ReLU(),
            nn.Dropout(0.20),
            nn.LayerNorm(dim_per_gene),
            Permute(0, 2, 1),  # B, G, H1 -> B, H1, G
            nn.Linear(5000, bottleneck_dim),  # B, H1, G -> B, H1, H2
            nn.ReLU(),
            nn.Dropout(0.20),
            nn.LayerNorm(bottleneck_dim),
            nn.Linear(bottleneck_dim, 5000),  # B, H1, H2 -> B, H1, 5000
            nn.ReLU(),
            nn.Dropout(0.20),
            nn.LayerNorm(5000),
            Permute(0, 2, 1),  # B, H1, 5000 -> B, 5000, H1
            nn.Linear(dim_per_gene, 1),  # B, 5000, 1
        )

        self.lr = float(self.config.lr)
        self.validation_step_outputs = []

    def forward(self, batch):
        x = self.to_embed(batch["gene_counts"].unsqueeze(-1))

        matches = torch.where((batch["gene_index"].T == batch["0/perturbed_gene"]).T)
        assert (matches[0] == torch.arange(len(x)).to(matches[0].device)).all()
        x[torch.arange(len(x)), matches[1]] += self.perturbation_indicator

        z = self.mixer(x).squeeze(-1)
        return z

    def training_step(self, batch, batch_idx):
        batch["gene_counts"] = batch["gene_counts"].to(self.dtype)

        y = self(batch)
        if self.config.pred_post_pert:
            target = batch["gene_counts_true"] - batch["gene_counts_copy"]
        else:
            target = batch["gene_counts_true"]

        loss = F.mse_loss(y, target.to(y.dtype))

        self.log("train_loss", loss, sync_dist=True)
        return loss

    def validation_step(self, batch, batch_idx):
        batch["gene_counts"] = batch["gene_counts"].to(self.dtype)

        y = self(batch)
        if self.config.pred_post_pert:
            target = batch["gene_counts_true"] - batch["gene_counts_copy"]
        else:
            target = batch["gene_counts_true"]

        loss = F.mse_loss(y, target.to(y.dtype))

        self.log("val_loss", loss, sync_dist=True)

        self.validation_step_outputs.append(
            (
                y.cpu(),
                batch["gene_counts_true"].cpu(),
                batch["gene_counts_copy"].cpu(),
                batch["0/perturbed_gene"].cpu(),
            )
        )

    def on_validation_epoch_end(self):
        pred_pert_prof = torch.cat([s[0] for s in self.validation_step_outputs])
        true_pert_prof = torch.cat([s[1] for s in self.validation_step_outputs])
        perturbed_genes = torch.cat([s[3] for s in self.validation_step_outputs])

        mean_ctrl_profile = torch.cat(
            [s[2] for s in self.validation_step_outputs]
        ).mean(0)

        pearsons = []
        delta_pearsons = []
        for pert_gene in torch.unique(perturbed_genes):
            pred_per_pert = pred_pert_prof[perturbed_genes == pert_gene].float().mean(0)
            true_per_pert = true_pert_prof[perturbed_genes == pert_gene].float().mean(0)

            pred_delta_per_pert = pred_per_pert - mean_ctrl_profile
            true_delta_per_pert = true_per_pert - mean_ctrl_profile

            s1 = pearsonr(pred_per_pert.numpy(), true_per_pert.numpy()).statistic
            pearsons.append(s1)

            s2 = pearsonr(
                pred_delta_per_pert.numpy(), true_delta_per_pert.numpy()
            ).statistic
            delta_pearsons.append(s2)
        self.log("val_pearson", np.mean(pearsons), sync_dist=True)
        self.log("val_deltapearson", np.mean(delta_pearsons), sync_dist=True)

        self.validation_step_outputs.clear()

    def predict_step(self, batch, batch_idx):
        batch["gene_counts"] = batch["gene_counts"].to(self.dtype)

        y = self(batch)
        return (
            y,
            batch["gene_counts_true"],
            batch["gene_counts_copy"],
            batch["gene_index"],
        )

    def configure_optimizers(self):
        optimizer = optim.Adam(self.parameters(), lr=self.lr)

        return optimizer

    @property
    def config_used(self):
        return {
            "baseline_perturb_dim_per_gene",
            "baseline_perturb_bottleneck_dim",
            "lr",
        }

    @property
    def config_unused(self):
        return set(self.config) - self.config_used


class CLSTaskBaseline(pl.LightningModule):
    def __init__(
        self,
        config,
    ):
        super().__init__()
        self.save_hyperparameters()

        self.config = deepcopy(config)
        layers = self.config.baseline_cls_task_layers

        if layers == []:
            self.net = nn.Identity()
            dim_to_loss = self.config.baseline_cls_hvg
        else:
            layers = [self.config.baseline_cls_hvg] + layers
            net = []
            for i in range(len(layers) - 1):
                net.append(nn.Linear(layers[i], layers[i + 1]))
                net.append(nn.Dropout(0.20))
                net.append(nn.LayerNorm(layers[i + 1]))
                net.append(nn.ReLU())
            self.net = nn.Sequential(*net)
            dim_to_loss = layers[-1]

        if self.config.celltype_clf_loss:
            self.loss = loss.CellTypeClfLoss(dim_to_loss, self.config.cls_finetune_dim)
        elif self.config.modality_prediction_loss:
            self.loss = loss.ModalityPredictionLoss(
                dim_to_loss, self.config.cls_finetune_dim
            )
        else:
            raise ValueError(
                "At least one of celltype clf loss or modality predict loss should be true"
            )

        self.lr = float(self.config.lr)
        self.validation_step_outputs = []

        if self.config.celltype_clf_loss:
            self.micro_acc = MulticlassAccuracy(
                num_classes=self.config.cls_finetune_dim, average="micro"
            )
            self.macro_acc = MulticlassAccuracy(
                num_classes=self.config.cls_finetune_dim, average="macro"
            )

    def forward(self, batch):
        z = self.net(batch["gene_counts"])
        return self.loss.predict(z)

    def training_step(self, batch, batch_idx):
        batch["gene_counts"] = batch["gene_counts"].to(self.dtype)

        y = self(batch)

        loss = self.loss.loss(y, batch["0/targets"])

        self.log("train_loss", loss, sync_dist=True)
        return loss

    def validation_step(self, batch, batch_idx):
        batch["gene_counts"] = batch["gene_counts"].to(self.dtype)

        y = self(batch)

        loss = self.loss.loss(y, batch["0/targets"])

        self.log("val_loss", loss, sync_dist=True)
        if self.config.celltype_clf_loss:
            self.micro_acc(y, batch["0/targets"])
            self.macro_acc(y, batch["0/targets"])
            self.log(
                "val_microacc",
                self.micro_acc,
                on_step=False,
                on_epoch=True,
                batch_size=len(batch["0/targets"]),
                sync_dist=True,
            )
            self.log(
                "val_macroacc",
                self.macro_acc,
                on_step=False,
                on_epoch=True,
                batch_size=len(batch["0/targets"]),
                sync_dist=True,
            )

        self.validation_step_outputs.append((y.cpu(), batch["0/targets"].cpu()))

    def on_validation_epoch_end(self):
        all_preds = torch.cat([s[0] for s in self.validation_step_outputs])
        all_trues = torch.cat([s[1] for s in self.validation_step_outputs])

        if isinstance(self.loss, loss.ModalityPredictionLoss):
            pearson_per_target = pearson_batch_masked(
                all_preds.T.float(), all_trues.T.float()
            ).numpy()
            self.log("val_macro_pearson", np.mean(pearson_per_target), sync_dist=True)

        self.validation_step_outputs.clear()

    def configure_optimizers(self):
        optimizer = optim.Adam(self.parameters(), lr=self.lr)

        return optimizer

    @property
    def config_used(self):
        return {
            "baseline_cls_task_layers",
            "celltype_clf_loss",
            "modality_prediction_loss",
            "cls_finetune_dim",
            "lr",
        }

    @property
    def config_unused(self):
        return set(self.config) - self.config_used


class VAE(pl.LightningModule):
    def __init__(
        self,
        config,
    ):
        super().__init__()
        self.save_hyperparameters()

        self.config = deepcopy(config)
        layers = self.config.vae_layers

        net = []
        for i in range(len(layers) - 1):
            net.append(nn.Linear(layers[i], layers[i + 1]))
            net.append(nn.Dropout(0.20))
            net.append(nn.LayerNorm(layers[i + 1]))
            net.append(nn.ReLU())
        self.encoder = nn.Sequential(*net)

        net = []
        for i in range(len(layers) - 2, 0, -1):
            net.append(nn.Linear(layers[i + 1], layers[i]))
            net.append(nn.Dropout(0.20))
            net.append(nn.LayerNorm(layers[i]))
            net.append(nn.ReLU())
        net.append(nn.Linear(layers[1], layers[0]))

        self.decoder = nn.Sequential(*net)

        self.to_mu = nn.Linear(layers[-1], layers[-1])
        self.to_var = nn.Linear(layers[-1], layers[-1])
        self.beta = self.config.beta

        self.lr = float(self.config.lr)

    def forward(self, batch):
        z = self.encoder(batch["gene_counts"])
        z_means, z_logvars = self.to_mu(z), self.to_var(z)

        z = self.reparameterization(z_means, z_logvars)
        x_reconstruct = self.decoder(z)

        return x_reconstruct, z_means, z_logvars

    def reparameterization(self, mean, logvar):
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        return mean + eps * std

    def KL_div(self, means, logvars):
        KL_div_loss = (
            -0.5 * torch.sum(1 + logvars - means**2 - logvars.exp())
        ) / means.size(1)
        return KL_div_loss

    def MSE(self, inputs, targets):
        return F.mse_loss(inputs, targets, reduction="none")

    def training_step(self, batch, batch_idx):
        batch["gene_counts"] = batch["gene_counts"].to(self.dtype)

        y, means, logvars = self(batch)
        pred_counts = y

        recon_loss = self.MSE(pred_counts, batch["gene_counts_true"])
        kl_div = self.KL_div(means, logvars)
        loss = (recon_loss + self.beta * kl_div).mean()

        self.log("train_loss", loss, sync_dist=True)
        return loss

    def validation_step(self, batch, batch_idx):
        batch["gene_counts"] = batch["gene_counts"].to(self.dtype)

        y, means, logvars = self(batch)
        pred_counts = y

        recon_loss = self.MSE(pred_counts, batch["gene_counts_true"])
        kl_div = self.KL_div(means, logvars)
        loss = (recon_loss + self.beta * kl_div).mean()

        self.log("val_loss", loss, sync_dist=True)
        return loss

    def predict_step(self, batch):
        batch["gene_counts"] = batch["gene_counts"].to(self.dtype)

        z = self.encoder(batch["gene_counts"])
        z_means = self.to_mu(z)

        y = self.decoder(z_means)

        count_predictions = torch.clamp(y.exp(), max=1e7)

        return (batch["0/obs"], z_means, count_predictions, batch["gene_counts_true"])

    def configure_optimizers(self):
        optimizer = optim.Adam(self.parameters(), lr=self.lr)
        return optimizer



import torch
from torch import optim, nn
import torch.nn.functional as F
import lightning.pytorch as pl
from bio_attention.attention import TransformerEncoder
from bio_attention.embed import DiscreteEmbedding, ContinuousEmbedding
from bento_sc import loss
from bento_sc.utils.metrics import pearson_batch_masked
from scipy.stats import pearsonr
from torchmetrics.classification import MulticlassAccuracy
import numpy as np
from copy import deepcopy


class EmbeddingGater(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.embedding = nn.Parameter(torch.empty(dim).uniform_(-1, 1))

    def forward(self, x):
        y = F.tanh(x) * self.embedding
        y[:, 0, :] = x[:, 0, :]
        return y


class EmbeddingPseudoQuantizer(nn.Module):
    def __init__(self, in_dim, out_dim):
        super().__init__()
        self.embedding = nn.Parameter(torch.empty(in_dim, out_dim).uniform_(-1, 1))

    def forward(self, x):
        y = F.softmax(x, dim=-1) @ self.embedding
        y[:, 0, :] = x[:, 0, :]
        return y


class BentoTransformer(pl.LightningModule):
    def __init__(self, config):
        super().__init__()
        self.save_hyperparameters()

        self.config = deepcopy(config)

        if self.config.discrete_input:
            self.embedder = DiscreteEmbedding(
                self.config.n_discrete_tokens, self.config.dim, cls=True
            )
        else:
            self.embedder = ContinuousEmbedding(self.config.dim, cls=True)

        if self.config.pseudoquant_input:
            self.embedder = nn.Sequential(
                self.embedder,
                EmbeddingPseudoQuantizer(self.config.dim, self.config.dim),
            )
        elif self.config.gate_input:
            self.embedder = nn.Sequential(
                self.embedder, EmbeddingGater(self.config.dim)
            )

        self.transformer = TransformerEncoder(
            depth=self.config.depth,
            dim=self.config.dim,
            nh=8,
            attentiontype="vanilla",
            attention_args={
                "dropout": self.config.dropout,
                "enable_math": False,
                "enable_flash": True,
                "enable_mem_efficient": True,
            },
            plugintype="learned",
            plugin_args={"dim": self.config.dim, "max_seq_len": self.config.n_genes},
            only_apply_plugin_at_first=True,
            dropout=self.config.dropout,
            glu_ff=True,
            activation="gelu",
        )

        loss_mapper = {
            "BinCE": loss.BinCE,
            "CountMSE": loss.CountMSE,
            "PoissonNLL": loss.PoissonNLL,
            "NegativeBinomialNLL": loss.NegativeBinomialNLL,
            "ZeroInflatedNegativeBinomialNLL": loss.ZeroInflatedNegativeBinomialNLL,
        }

        loss_type = self.config.loss["type"]
        loss_kwargs = {k: v for k, v in self.config.loss.items() if k != "type"}
        self.loss = loss_mapper[loss_type](self.config.dim, **loss_kwargs)

        if self.config.nce_loss:
            self.nce_loss = loss.NCELoss(
                self.config.dim, self.config.nce_dim, temperature=self.config.nce_temp
            )

        if self.config.celltype_clf_loss:
            self.ct_clf_loss = loss.CellTypeClfLoss(self.config.dim, 164)

        self.lr = float(self.config.lr)

    def forward(self, batch):
        mask = batch["gene_counts"] != -1
        if torch.all(mask):
            mask = None
        x = self.embedder(batch["gene_counts"])
        z = self.transformer(x, pos=batch["gene_index"], mask=mask)
        return z

    def training_step(self, batch, batch_idx):
        if not self.config.discrete_input:
            batch["gene_counts"] = batch["gene_counts"].to(self.dtype)
        else:
            batch["gene_counts"] = batch["gene_counts"].float()

        if not self.config.train_on_all:
            train_on = torch.isnan(batch["gene_counts"])
        else:
            train_on = None

        y = self(batch)

        if ("no_genewise_loss" in self.config) and (
            self.config["no_genewise_loss"] == True
        ):
            loss = 0
        else:
            loss = self.loss(
                y[:, 1:],
                batch["gene_counts_true"],
                gene_ids=batch["gene_index"],
                train_on=train_on,
            )

        if self.config.nce_loss:
            nce_loss = self.nce_loss(y[:, 0])
            loss += nce_loss

        if self.config.celltype_clf_loss:
            ct_loss = self.ct_clf_loss(y[:, 0], batch["0/obs"][:, 3])
            loss += ct_loss

        self.log("train_loss", loss, sync_dist=True)
        return loss

    def validation_step(self, batch, batch_idx):
        if not self.config.discrete_input:
            batch["gene_counts"] = batch["gene_counts"].to(self.dtype)
        else:
            batch["gene_counts"] = batch["gene_counts"].float()

        if not self.config.train_on_all:
            train_on = torch.isnan(batch["gene_counts"])
        else:
            train_on = None

        y = self(batch)

        if ("no_genewise_loss" in self.config) and (
            self.config["no_genewise_loss"] == True
        ):
            loss = 0
        else:
            loss = self.loss(
                y[:, 1:],
                batch["gene_counts_true"],
                gene_ids=batch["gene_index"],
                train_on=train_on,
            )

        if self.config.nce_loss:
            nce_loss = self.nce_loss(y[:, 0])
            loss += nce_loss

        if self.config.celltype_clf_loss:
            ct_loss = self.ct_clf_loss(y[:, 0], batch["0/obs"][:, 3])
            loss += ct_loss

        self.log("val_loss", loss, sync_dist=True)

    def predict_step(self, batch):
        if not self.config.discrete_input:
            batch["gene_counts"] = batch["gene_counts"].to(self.dtype)
        else:
            batch["gene_counts"] = batch["gene_counts"].float()

        y = self(batch)
        libsizes = (batch["gene_counts"].sum(1) + (batch["gene_counts"] == -1).sum(1))[
            :, None
        ]

        count_predictions = self.loss.predict(
            y[:, 1:],
            gene_ids=batch["gene_index"],
            libsize=libsizes,
        )
        if isinstance(count_predictions, tuple):
            count_predictions = count_predictions[0]

        return (batch["0/obs"], y[:, 0], count_predictions, batch["gene_counts_true"])

    def configure_optimizers(self):
        optimizer = optim.Adam(self.parameters(), lr=self.lr)

        return optimizer

    @property
    def config_used(self):
        return {
            "n_discrete_tokens",
            "dim",
            "pseudoquant_input",
            "gate_input",
            "depth",
            "dropout",
            "n_genes",
            "loss",
            "nce_loss",
            "nce_dim",
            "nce_temp",
            "celltype_clf_loss",
            "lr",
            "train_on_all",
        }

    @property
    def config_unused(self):
        return set(self.config) - self.config_used


class PerturbTransformer(BentoTransformer):
    def __init__(self, config):
        super().__init__(config)
        assert self.config.nce_loss == False
        assert self.config.celltype_clf_loss == False
        assert self.config.train_on_all == True
        assert isinstance(self.loss, loss.CountMSE)

        self.perturbation_indicator = nn.Parameter(
            torch.empty(self.config.dim).uniform_(-1, 1)
            / self.config.perturb_init_factor
        )
        self.validation_step_outputs = []

    def forward(self, batch):
        mask = batch["gene_counts"] != -1
        if torch.all(mask):
            mask = None

        x = self.embedder(batch["gene_counts"])

        matches = torch.where((batch["gene_index"].T == batch["0/perturbed_gene"]).T)
        assert (matches[0] == torch.arange(len(x)).to(matches[0].device)).all()
        x[torch.arange(len(x)), matches[1]] += self.perturbation_indicator

        z = self.transformer(x, pos=batch["gene_index"], mask=mask)
        return z

    def training_step(self, batch, batch_idx):
        if not self.config.discrete_input:
            batch["gene_counts"] = batch["gene_counts"].to(self.dtype)
        else:
            batch["gene_counts"] = batch["gene_counts"].float()

        y = self(batch)

        loss = self.loss(
            y[:, 1:],
            batch["gene_counts_true"].to(y.dtype),
            gene_ids=batch["gene_index"],
        )

        self.log("train_loss", loss, sync_dist=True)
        return loss

    def validation_step(self, batch, batch_idx):
        if not self.config.discrete_input:
            batch["gene_counts"] = batch["gene_counts"].to(self.dtype)
        else:
            batch["gene_counts"] = batch["gene_counts"].float()

        y = self(batch)

        loss = self.loss(
            y[:, 1:],
            batch["gene_counts_true"].to(y.dtype),
            gene_ids=batch["gene_index"],
        )

        self.log("val_loss", loss, sync_dist=True)

        libsizes = batch["gene_counts_true"].sum(1) + (
            batch["gene_counts_true"] == -1
        ).sum(1)

        y = self.loss.predict(y[:, 1:], libsize=libsizes)

        self.validation_step_outputs.append(
            (
                y.cpu(),
                batch["gene_counts_true"].cpu(),
                batch["gene_counts_copy"].cpu(),
                batch["0/perturbed_gene"].cpu(),
            )
        )

    def on_validation_epoch_end(self):
        pred_pert_prof = torch.cat([s[0] for s in self.validation_step_outputs])
        true_pert_prof = torch.cat([s[1] for s in self.validation_step_outputs])
        perturbed_genes = torch.cat([s[3] for s in self.validation_step_outputs])

        mean_ctrl_profile = torch.cat(
            [s[2] for s in self.validation_step_outputs]
        ).mean(0)

        pearsons = []
        delta_pearsons = []
        for pert_gene in torch.unique(perturbed_genes):
            pred_per_pert = pred_pert_prof[perturbed_genes == pert_gene].float().mean(0)
            true_per_pert = true_pert_prof[perturbed_genes == pert_gene].float().mean(0)

            pred_delta_per_pert = pred_per_pert - mean_ctrl_profile
            true_delta_per_pert = true_per_pert - mean_ctrl_profile

            s1 = pearsonr(pred_per_pert.numpy(), true_per_pert.numpy()).statistic
            pearsons.append(s1)

            s2 = pearsonr(
                pred_delta_per_pert.numpy(), true_delta_per_pert.numpy()
            ).statistic
            delta_pearsons.append(s2)
        self.log("val_pearson", np.mean(pearsons), sync_dist=True)
        self.log("val_deltapearson", np.mean(delta_pearsons), sync_dist=True)

        self.validation_step_outputs.clear()

    def predict_step(self, batch, batch_idx):
        if not self.config.discrete_input:
            batch["gene_counts"] = batch["gene_counts"].to(self.dtype)
        else:
            batch["gene_counts"] = batch["gene_counts"].float()

        y = self(batch)

        libsizes = batch["gene_counts_true"].sum(1) + (
            batch["gene_counts_true"] == -1
        ).sum(1)
        y = self.loss.predict(y[:, 1:], libsize=libsizes)
        return (
            y,
            batch["gene_counts_true"],
            batch["gene_counts_copy"],
            batch["gene_index"],
        )

    @property
    def config_used(self):
        return {
            "n_discrete_tokens",
            "dim",
            "pseudoquant_input",
            "gate_input",
            "depth",
            "dropout",
            "n_genes",
            "loss",
            "lr",
        }


class CLSTaskTransformer(BentoTransformer):
    def __init__(self, config):
        super().__init__(config)
        assert self.config.nce_loss == False

        if self.config.celltype_clf_loss:
            self.loss = loss.CellTypeClfLoss(
                self.config.dim, self.config.cls_finetune_dim
            )
        elif self.config.modality_prediction_loss:
            self.loss = loss.ModalityPredictionLoss(
                self.config.dim, self.config.cls_finetune_dim
            )
        else:
            raise ValueError(
                "At least one of celltype clf loss or modality predict loss should be true"
            )

        self.validation_step_outputs = []

        if self.config.celltype_clf_loss:
            self.micro_acc = MulticlassAccuracy(
                num_classes=self.config.cls_finetune_dim, average="micro"
            )
            self.macro_acc = MulticlassAccuracy(
                num_classes=self.config.cls_finetune_dim, average="macro"
            )

    def forward(self, batch):
        mask = batch["gene_counts"] != -1
        if torch.all(mask):
            mask = None
        x = self.embedder(batch["gene_counts"])
        z = self.transformer(x, pos=batch["gene_index"], mask=mask)
        return self.loss.predict(z[:, 0])

    def training_step(self, batch, batch_idx):
        if not self.config.discrete_input:
            batch["gene_counts"] = batch["gene_counts"].to(self.dtype)
        else:
            batch["gene_counts"] = batch["gene_counts"].float()

        y = self(batch)

        loss = self.loss.loss(y, batch["0/targets"])

        self.log("train_loss", loss, sync_dist=True)
        return loss

    def validation_step(self, batch, batch_idx):
        if not self.config.discrete_input:
            batch["gene_counts"] = batch["gene_counts"].to(self.dtype)
        else:
            batch["gene_counts"] = batch["gene_counts"].float()

        y = self(batch)

        loss = self.loss.loss(y, batch["0/targets"])

        self.log("val_loss", loss, sync_dist=True)
        if self.config.celltype_clf_loss:
            self.micro_acc(y, batch["0/targets"])
            self.macro_acc(y, batch["0/targets"])
            self.log(
                "val_microacc",
                self.micro_acc,
                on_step=False,
                on_epoch=True,
                batch_size=len(batch["0/targets"]),
                sync_dist=True,
            )
            self.log(
                "val_macroacc",
                self.macro_acc,
                on_step=False,
                on_epoch=True,
                batch_size=len(batch["0/targets"]),
                sync_dist=True,
            )

        self.validation_step_outputs.append((y.cpu(), batch["0/targets"].cpu()))

    def on_validation_epoch_end(self):
        all_preds = torch.cat([s[0] for s in self.validation_step_outputs])
        all_trues = torch.cat([s[1] for s in self.validation_step_outputs])

        if isinstance(self.loss, loss.ModalityPredictionLoss):
            pearson_per_target = pearson_batch_masked(
                all_preds.T.float(), all_trues.T.float()
            ).numpy()
            self.log("val_macro_pearson", np.mean(pearson_per_target), sync_dist=True)

        self.validation_step_outputs.clear()

    @property
    def config_used(self):
        return {
            "n_discrete_tokens",
            "dim",
            "pseudoquant_input",
            "gate_input",
            "depth",
            "dropout",
            "n_genes",
            "lr",
            "celltype_clf_loss",
            "modality_prediction_loss",
            "cls_finetune_dim",
        }



def pearson_batch_masked(preds, trues):

    mask = trues == -1

    preds = preds.clone()
    trues = trues.clone()

    preds[mask] = 0
    trues[mask] = 0

    preds_mean = preds.sum(1) / (~mask).sum(1)
    trues_mean = trues.sum(1) / (~mask).sum(1)

    pred_min_mean = preds - preds_mean[:, None]
    true_min_mean = trues - trues_mean[:, None]

    pred_min_mean[mask] = 0
    true_min_mean[mask] = 0

    numer = (pred_min_mean * true_min_mean).sum(1)
    denom = ((pred_min_mean**2).sum(1) * (true_min_mean**2).sum(1)) ** 0.5

    return numer / denom



import yaml


class Config(dict):
    def __init__(self, path=None, **kwargs):
        if isinstance(path, str):
            with open(path) as f:
                path = yaml.load(f, Loader=yaml.Loader)
            super().__init__(path)
        else:
            super().__init__(**kwargs)

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def write(self, path):
        with open(path, "w") as outfile:
            yaml.dump(dict(self), outfile, default_flow_style=False, sort_keys=False)

    def change_keys(self, **kwargs):
        for k, v in kwargs.items():
            self[k] = v

    def print_used_keys(self, used_keys):
        unused_keys = set(self) - used_keys
        print("Config used:")
        for i in used_keys:
            name = (i[:18] + "..") if len(i) > 20 else i
            print(name[:20] + " " * (20 - len(i)), "\t", self[i])
        print("Config unused:")
        for i in unused_keys:
            name = (i[:18] + "..") if len(i) > 20 else i
            print(name[:20] + " " * (20 - len(i)), "\t", self[i])



from os.path import join
import pandas as pd
import numpy as np
from tqdm.auto import tqdm
import obonet
import networkx
from importlib.resources import files
import cellxgene_census
import argparse
import os
import anndata
from math import ceil
import h5py
import sys
from scipy.sparse import csr_matrix
import urllib.request
import gzip
import shutil
import h5torch
from bento_sc.data import CellSampleProcessor, SequentialPreprocessor, FilterTopGenes


class CustomFormatter(
    argparse.ArgumentDefaultsHelpFormatter, argparse.MetavarTypeHelpFormatter
):
    pass


def main():
    parser = argparse.ArgumentParser(
        description="Data downloading script launching pad. Choose a datafile to download and process to h5torch format.",
        formatter_class=CustomFormatter,
    )

    parser.add_argument(
        "datafile",
        type=str,
        metavar="datafile",
        choices=[
            "scTab",
            "perturb",
            "neurips_citeseq",
            "replogle_perturb",
            "scTab_upscaling",
            "sctab_grn",
        ],
        help="Datafile to download, choices: {%(choices)s}",
    )

    args = parser.parse_args(sys.argv[1:2])

    mapper = {
        "scTab": (get_sctab_parser, get_sctab),
        "scTab_upscaling": (get_sctab_upsc_parser, get_sctab_upsc),
        "scTab_grn": (get_sctab_grn_parser, get_sctab_grn),
        "neurips_citeseq": (get_citeseq_parser, get_citeseq),
        "replogle_perturb": (get_perturb_parser, get_perturb),
        "batchcorr_embryolimb": (get_el_parser, get_el),
        "batchcorr_greatapes": (get_ga_parser, get_ga),
        "batchcorr_circimm": (get_ci_parser, get_ci),
    }

    parser_func, download_func = mapper[args.datafile]
    parser = parser_func()
    args = parser.parse_args(sys.argv[2:])
    download_func(args)


def get_sctab_parser():
    parser = argparse.ArgumentParser(
        description="Download and format scTab.",
        formatter_class=CustomFormatter,
    )
    parser.add_argument(
        "base_folder",
        type=str,
        metavar="base_folder",
        help="base folder to put intermediate files in (will occupy +-256gb). Note that this script does not delete intermediate files.",
    )
    parser.add_argument(
        "output_file_path",
        type=str,
        metavar="output_file_path",
        help="output .h5t file to write processed data to.",
    )
    return parser


def get_sctab_upsc_parser():
    parser = argparse.ArgumentParser(
        description="Subset scTab for gene upscaling task (select deeply sequenced cells).",
        formatter_class=CustomFormatter,
    )
    parser.add_argument(
        "sctab_h5t",
        type=str,
        metavar="sctab_h5t",
        help="path to processed sctab h5t file.",
    )
    parser.add_argument(
        "output_file_path",
        type=str,
        metavar="output_file_path",
        help="output .h5t file to write processed data to.",
    )
    parser.add_argument(
        "val_or_test",
        type=str,
        choices=["val", "test"],
        metavar="val_or_test",
        help="Data split to subset. Choices: {%(choices)s}",
    )
    return parser


def get_sctab_grn_parser():
    parser = argparse.ArgumentParser(
        description="Subset scTab for GRN inference task (i.e. select 2~500 cells from 4 celltypes in a given data split).",
        formatter_class=CustomFormatter,
    )
    parser.add_argument(
        "sctab_h5t",
        type=str,
        metavar="sctab_h5t",
        help="path to processed sctab h5t file.",
    )
    parser.add_argument(
        "output_file_path",
        type=str,
        metavar="output_file_path",
        help="output .h5t file to write processed data to.",
    )
    parser.add_argument(
        "val_or_test",
        type=str,
        choices=["val", "test"],
        metavar="val_or_test",
        help="Data split to subset. Choices: {%(choices)s}",
    )
    return parser


def get_citeseq_parser():
    parser = argparse.ArgumentParser(
        description="Download and format NeurIPS 2021 CITE-seq dataset.",
        formatter_class=CustomFormatter,
    )
    parser.add_argument(
        "base_folder",
        type=str,
        metavar="base_folder",
        help="base folder to put intermediate files in. Note that this script does not delete intermediate files.",
    )
    parser.add_argument(
        "sctab_h5t",
        type=str,
        metavar="sctab_h5t",
        help="path to processed sctab h5t file. Used to unify gene vocabulary.",
    )
    parser.add_argument(
        "output_file_path",
        type=str,
        metavar="output_file_path",
        help="output .h5t file to write processed data to.",
    )
    return parser


def get_perturb_parser():
    parser = argparse.ArgumentParser(
        description="Download and format Replogle Perturb-seq dataset.",
        formatter_class=CustomFormatter,
    )
    parser.add_argument(
        "base_folder",
        type=str,
        metavar="base_folder",
        help="base folder to put intermediate files in. Note that this script does not delete intermediate files.",
    )
    parser.add_argument(
        "sctab_h5t",
        type=str,
        metavar="sctab_h5t",
        help="path to processed sctab h5t file. Used to unify gene vocabulary.",
    )
    parser.add_argument(
        "output_file_path",
        type=str,
        metavar="output_file_path",
        help="output .h5t file to write processed data to.",
    )
    return parser


def get_el_parser():
    parser = argparse.ArgumentParser(
        description="Download and format Embryonic Limb dataset.",
        formatter_class=CustomFormatter,
    )
    parser.add_argument(
        "base_folder",
        type=str,
        metavar="base_folder",
        help="base folder to put intermediate files in. Note that this script does not delete intermediate files.",
    )
    parser.add_argument(
        "sctab_h5t",
        type=str,
        metavar="sctab_h5t",
        help="path to processed sctab h5t file. Used to unify gene vocabulary.",
    )
    parser.add_argument(
        "output_file_path",
        type=str,
        metavar="output_file_path",
        help="output .h5t file to write processed data to.",
    )
    return parser


def get_ga_parser():
    parser = argparse.ArgumentParser(
        description="Download and Great Apes dataset.",
        formatter_class=CustomFormatter,
    )
    parser.add_argument(
        "base_folder",
        type=str,
        metavar="base_folder",
        help="base folder to put intermediate files in. Note that this script does not delete intermediate files.",
    )
    parser.add_argument(
        "sctab_h5t",
        type=str,
        metavar="sctab_h5t",
        help="path to processed sctab h5t file. Used to unify gene vocabulary.",
    )
    parser.add_argument(
        "output_file_path",
        type=str,
        metavar="output_file_path",
        help="output .h5t file to write processed data to.",
    )
    return parser


def get_ci_parser():
    parser = argparse.ArgumentParser(
        description="Download and format Circulating Immune cells dataset.",
        formatter_class=CustomFormatter,
    )
    parser.add_argument(
        "base_folder",
        type=str,
        metavar="base_folder",
        help="base folder to put intermediate files in. Note that this script does not delete intermediate files.",
    )
    parser.add_argument(
        "sctab_h5t",
        type=str,
        metavar="sctab_h5t",
        help="path to processed sctab h5t file. Used to unify gene vocabulary.",
    )
    parser.add_argument(
        "output_file_path",
        type=str,
        metavar="output_file_path",
        help="output .h5t file to write processed data to.",
    )
    return parser


def get_sctab(args):
    # in large part adapted from https://github.com/theislab/scTab,
    # with some changes to make it work for limited RAM systems.
    url = "https://github.com/obophenotype/cell-ontology/releases/download/v2023-05-22/cl-simple.obo"
    graph = obonet.read_obo(url, ignore_obsolete=True)

    # only use "is_a" edges
    edges_to_delete = []
    for i, x in enumerate(graph.edges):
        if x[2] != "is_a":
            edges_to_delete.append((x[0], x[1]))
    for x in edges_to_delete:
        graph.remove_edge(u=x[0], v=x[1])

    # define mapping from id to name
    id_to_name = {id_: data.get("name") for id_, data in graph.nodes(data=True)}
    # define inverse mapping from name to id
    name_to_id = {v: k for k, v in id_to_name.items()}

    def find_child_nodes(cell_type):
        return [
            id_to_name[node]
            for node in networkx.ancestors(graph, name_to_id[cell_type])
        ]

    def find_parent_nodes(cell_type):
        return [
            id_to_name[node]
            for node in networkx.descendants(graph, name_to_id[cell_type])
        ]

    census = cellxgene_census.open_soma(census_version="2023-05-15")

    summary = census["census_info"]["summary"]

    PROTOCOLS = [
        "10x 5' v2",
        "10x 3' v3",
        "10x 3' v2",
        "10x 5' v1",
        "10x 3' v1",
        "10x 3' transcription profiling",
        "10x 5' transcription profiling",
    ]

    COLUMN_NAMES = [
        "soma_joinid",
        "is_primary_data",
        "dataset_id",
        "donor_id",
        "assay",
        "cell_type",
        "development_stage",
        "disease",
        "tissue",
        "tissue_general",
    ]

    obs = (
        census["census_data"]["homo_sapiens"]
        .obs.read(
            column_names=COLUMN_NAMES,
            value_filter=f"is_primary_data == True and assay in {PROTOCOLS}",
        )
        .concat()
        .to_pandas()
    )

    obs["tech_sample"] = (obs.dataset_id + "_" + obs.donor_id).astype("category")

    for col in COLUMN_NAMES:
        if obs[col].dtype == object:
            obs[col] = obs[col].astype("category")

    # remove all cell types which are not a subtype of native cell
    cell_types_to_remove = (
        obs[~obs.cell_type.isin(find_child_nodes("native cell"))]
        .cell_type.unique()
        .tolist()
    )

    # remove all cell types which have less than 5000 cells
    cell_freq = obs.cell_type.value_counts()
    cell_types_to_remove += cell_freq[cell_freq < 5000].index.tolist()

    # remove cell types which have less than 30 tech_samples
    tech_samples_per_cell_type = (
        obs[["cell_type", "tech_sample"]]
        .groupby("cell_type")
        .agg({"tech_sample": "nunique"})
        .sort_values("tech_sample")
    )
    cell_types_to_remove += tech_samples_per_cell_type[
        tech_samples_per_cell_type.tech_sample <= 30
    ].index.tolist()

    # filter out too granular labels
    # remove all cells that have <= 7 parents in the cell ontology
    cell_types = obs.cell_type.unique().tolist()

    n_children = []
    n_parents = []

    for cell_type in cell_types:
        n_parents.append(len(find_parent_nodes(cell_type)))
        n_children.append(len(find_child_nodes(cell_type)))

    cell_types_to_remove += (
        pd.DataFrame(
            {"n_children": n_children, "n_parents": n_parents}, index=cell_types
        )
        .query("n_parents <= 7")
        .index.tolist()
    )
    cell_types_to_remove = list(set(cell_types_to_remove))

    obs_subset = obs[~obs.cell_type.isin(cell_types_to_remove)].copy()
    for col in obs_subset.columns:
        if obs_subset[col].dtype == "category":
            obs_subset[col] = obs_subset[col].cat.remove_unused_categories()

    protein_coding_genes = pd.read_parquet(
        "https://raw.githubusercontent.com/theislab/scTab/devel/notebooks/store_creation/features.parquet"
    ).gene_names.tolist()

    # download in batches to not run out of memory
    for i, idxs in tqdm(
        enumerate(np.array_split(obs_subset.soma_joinid.to_numpy(), 100))
    ):
        adata = cellxgene_census.get_anndata(
            census=census,
            organism="Homo sapiens",
            X_name="raw",
            obs_coords=idxs.tolist(),
            var_value_filter=f"feature_name in {protein_coding_genes}",
            column_names={"obs": COLUMN_NAMES, "var": ["feature_id", "feature_name"]},
        )
        adata.write_h5ad(join(args.base_folder, f"{i}.h5ad"))

    files = [
        join(args.base_folder, file)
        for file in sorted(
            os.listdir(args.base_folder), key=lambda x: int(x.split(".")[0])
        )
        if file.endswith(".h5ad")
    ]

    def read_obs(path):
        obs = anndata.read_h5ad(path, backed="r").obs
        obs["tech_sample"] = obs.dataset_id.astype(str) + "_" + obs.donor_id.astype(str)
        return obs

    # read obs
    print("Loading obs...")
    obs = pd.concat([read_obs(file) for file in files]).reset_index(drop=True)
    for col in obs.columns:
        if obs[col].dtype == object:
            obs[col] = obs[col].astype("category")
            obs[col].cat.remove_unused_categories()

    def get_split(samples, val_split: float = 0.15, test_split: float = 0.15, seed=1):
        rng = np.random.default_rng(seed=seed)

        samples = np.array(samples)
        rng.shuffle(samples)
        n_samples = len(samples)

        n_samples_val = ceil(val_split * n_samples)
        n_samples_test = ceil(test_split * n_samples)
        n_samples_train = n_samples - n_samples_val - n_samples_test

        return {
            "train": samples[:n_samples_train],
            "val": samples[n_samples_train : (n_samples_train + n_samples_val)],
            "test": samples[(n_samples_train + n_samples_val) :],
        }

    def subset(splits, frac):
        assert 0.0 < frac <= 1.0
        if frac == 1.0:
            return splits
        else:
            return splits[: ceil(frac * len(splits))]

    splits = {"train": [], "val": [], "test": []}
    tech_sample_splits = get_split(obs.tech_sample.unique().tolist())
    for x in ["train", "val", "test"]:
        # tech_samples are already shuffled in the get_split method -> just subselect to subsample donors
        if x == "train":
            # only subset training data set
            splits[x] = obs[
                obs.tech_sample.isin(subset(tech_sample_splits[x], 1.0))
            ].index.to_numpy()
        else:
            splits[x] = obs[
                obs.tech_sample.isin(tech_sample_splits[x])
            ].index.to_numpy()

    assert len(np.intersect1d(splits["train"], splits["val"])) == 0
    assert len(np.intersect1d(splits["train"], splits["test"])) == 0
    assert len(np.intersect1d(splits["val"], splits["train"])) == 0
    assert len(np.intersect1d(splits["val"], splits["test"])) == 0

    rng = np.random.default_rng(seed=1)

    splits["train"] = rng.permutation(splits["train"])
    splits["val"] = rng.permutation(splits["val"])
    splits["test"] = rng.permutation(splits["test"])

    splits2 = {}
    splits2["train"] = splits["train"][: (len(splits["train"]) // 1024) * 1024]
    splits2["val"] = splits["val"][: (len(splits["val"]) // 1024) * 1024]
    splits2["test"] = splits["test"][: (len(splits["test"]) // 1024) * 1024]

    len_ = len(obs)

    f_out = h5torch.File(args.output_file_path, "w")

    f = h5py.File(os.path.join(args.base_folder, "0.h5ad"))
    mat = csr_matrix(
        (f["X/data"], f["X/indices"], f["X/indptr"].astype(np.int64)),
        shape=(f["X/indptr"].shape[0] - 1, 19331),
    )
    mat.indices = mat.indices.astype("int16")
    mat.indptr = mat.indptr.astype("int64")
    f_out.register(
        mat,
        axis="central",
        mode="csr",
        dtype_save="float32",
        dtype_load="float32",
        csr_load_sparse=True,
        length=len_,
    )
    f.close()

    for i in tqdm(range(1, 100)):
        f = h5py.File(os.path.join(args.base_folder, "%s.h5ad" % i))
        mat = csr_matrix(
            (f["X/data"], f["X/indices"], f["X/indptr"]),
            shape=(f["X/indptr"].shape[0] - 1, 19331),
        )
        mat.indices = mat.indices.astype("int16")
        mat.indptr = mat.indptr.astype("int64")
        f_out.append(mat, "central")
        f.close()

    f_out.close()

    f_out = h5torch.File(args.output_file_path, "a")

    obs_ = np.stack([obs[i].cat.codes.values for i in obs.columns[2:]]).T
    categories = {
        str(i) + "_" + k: obs[k].cat.categories.values.astype(bytes)
        for i, k in enumerate(obs.columns[2:])
    }

    f_out.register(obs_, axis=0, name="obs", dtype_load="int64")

    for k, v in categories.items():
        f_out.register(
            v, axis="unstructured", name=k, dtype_save="bytes", dtype_load="str"
        )

    def read_var(path):
        return anndata.read_h5ad(path, backed="r").var

    var = read_var(files[0])
    f_out.register(
        var.values.astype(bytes),
        axis=1,
        name="var",
        dtype_save="bytes",
        dtype_load="str",
    )
    f_out.close()

    import h5torch

    split_h5 = np.full(len(obs), "NA", dtype=object)
    split_h5[splits2["train"]] = "train"
    split_h5[splits2["val"]] = "val"
    split_h5[splits2["test"]] = "test"

    val_sub = np.random.permutation(splits2["val"])[:100_000]
    split_h5[val_sub] = "val_sub"

    f_out = h5torch.File(args.output_file_path, "a")
    f_out.register(
        split_h5.astype(bytes),
        axis=0,
        name="split",
        dtype_save="bytes",
        dtype_load="str",
    )
    f_out.close()
    return None


def get_sctab_upsc(args):
    d = h5torch.Dataset(
        args.sctab_h5t,
        sample_processor=CellSampleProcessor(
            SequentialPreprocessor(), return_zeros=False
        ),
        subset=("0/split", args.val_or_test),
    )
    k = np.load(
        files("bento_sc.utils.data").joinpath(
            "cxg_upsc_%s_subset.npy" % args.val_or_test
        )
    )

    matrix = np.zeros((25_000, 19331), dtype="int32")
    from scipy.sparse import csr_matrix

    obs_ = []
    for ix, n in tqdm(enumerate(k)):
        matrix[ix, d[n]["gene_index"].numpy()] = (
            d[n]["gene_counts"].numpy().astype(np.int32)
        )
        obs_.append(d[n]["0/obs"])

    obs = np.stack(obs_)
    f_out = h5torch.File(args.output_file_path, "w")
    f_out.register(
        csr_matrix(matrix),
        axis="central",
        mode="csr",
        dtype_save="float32",
        dtype_load="float32",
        csr_load_sparse=True,
    )
    f_out.register(obs, axis=0, name="obs", dtype_save="int64", dtype_load="int64")

    f_out.register(
        d.f["1/var"][:],
        axis=1,
        name="var",
        dtype_save="bytes",
        dtype_load="str",
    )

    f_out.register(
        k,
        axis="unstructured",
        name="samples",
    )

    split = np.full((25_000), "test")
    f_out.register(split, axis=0, name="split", dtype_save="bytes", dtype_load="str")

    f_out.close()
    return None


def get_sctab_grn(args):
    url = "https://github.com/obophenotype/cell-ontology/releases/download/v2023-05-22/cl-simple.obo"
    graph = obonet.read_obo(url, ignore_obsolete=True)

    # only use "is_a" edges
    edges_to_delete = []
    for i, x in enumerate(graph.edges):
        if x[2] != "is_a":
            edges_to_delete.append((x[0], x[1]))
    for x in edges_to_delete:
        graph.remove_edge(u=x[0], v=x[1])

    # define mapping from id to name
    id_to_name = {id_: data.get("name") for id_, data in graph.nodes(data=True)}
    # define inverse mapping from name to id
    name_to_id = {v: k for k, v in id_to_name.items()}

    def find_child_nodes(cell_type):
        return [
            id_to_name[node]
            for node in networkx.ancestors(graph, name_to_id[cell_type])
        ]

    myeloid_cells = ["myeloid cell"] + find_child_nodes("myeloid cell")
    B_cells = ["B cell"] + find_child_nodes("B cell")
    NK_cells = ["natural killer cell"] + find_child_nodes("natural killer cell")
    T_cells = ["T cell"] + find_child_nodes("T cell")

    f_cxg = h5torch.File(args.sctab_h5t)
    ct_cxg = f_cxg["unstructured/3_cell_type"][:].astype(str)

    celltypes_indices = []
    for celltype_to_select in [myeloid_cells, B_cells, NK_cells, T_cells]:
        indices_celltype = np.where(
            np.isin(ct_cxg, np.array([ct for ct in celltype_to_select if ct in ct_cxg]))
        )[0]
        celltypes_indices.append(np.isin(f_cxg["0/obs"][:, 3], indices_celltype))

    blood_cells = (
        f_cxg["0/obs"][:, 7]
        == np.where(f_cxg["unstructured"]["7_tissue_general"][:] == b"blood")[0][0]
    )
    celltype_indices_blood = []
    for celltype_ind in celltypes_indices:
        celltype_indices_blood.append(np.logical_and(celltype_ind, blood_cells))

    frac = f_cxg["0/split"][:] == bytes(args.val_or_test, "utf-8")
    celltype_indices_blood_frac = []
    for celltype_ind in celltype_indices_blood:
        celltype_indices_blood_frac.append(np.logical_and(celltype_ind, frac))

    matrix = np.zeros((10_000, 19331), dtype="int32")
    obs_ = []

    subsets = np.load(
        files("bento_sc.utils.data").joinpath(
            "cxg_grn_%s_subset.npy" % args.val_or_test
        )
    )

    c = 0
    for subset, ct_ind in zip(np.split(subsets, 4), celltype_indices_blood_frac):
        d = h5torch.Dataset(
            args.sctab_h5t,
            sample_processor=CellSampleProcessor(
                SequentialPreprocessor(
                    FilterTopGenes(
                        affected_keys=["gene_counts", "gene_index", "gene_counts_true"],
                        number=1024,
                    )
                ),
                return_zeros=False,
            ),
            subset=np.where(ct_ind)[0],
        )

        for n in tqdm(subset):
            matrix[c, d[n]["gene_index"].numpy()] = (
                d[n]["gene_counts"].numpy().astype(np.int32)
            )
            c += 1
            obs_.append(d[n]["0/obs"])

    celltype = np.array(
        ["Myeloid cells"] * 2500
        + ["B cells"] * 2500
        + ["NK cells"] * 2500
        + ["T cells"] * 2500
    ).astype(bytes)

    obs = np.stack(obs_)

    f_out = h5torch.File(args.output_file_path, "w")
    f_out.register(
        csr_matrix(matrix),
        axis="central",
        mode="csr",
        dtype_save="float32",
        dtype_load="float32",
        csr_load_sparse=True,
    )
    f_out.register(obs, axis=0, name="obs", dtype_save="int64", dtype_load="int64")

    f_out.register(
        celltype, axis=0, name="celltype", dtype_save="bytes", dtype_load="str"
    )

    f_out.register(
        d.f["1/var"][:],
        axis=1,
        name="var",
        dtype_save="bytes",
        dtype_load="str",
    )

    f_out.register(
        np.concatenate(subsets),
        axis="unstructured",
        name="samples",
    )

    split = np.full((10_000), "test")
    f_out.register(split, axis=0, name="split", dtype_save="bytes", dtype_load="str")

    f_out.close()
    return None


def get_perturb(args):
    urllib.request.urlretrieve(
        "https://plus.figshare.com/ndownloader/files/35773219",
        os.path.join(args.base_folder, "./perturb.h5ad"),
    )

    f = h5py.File(os.path.join(args.base_folder, "./perturb.h5ad"))

    gene_ids_pert = f["var/gene_id"][:]
    f_cxg = h5py.File(args.sctab_h5t)
    gene_ids_cxg = f_cxg["1/var"][:, 0]

    indices_map = []
    for g_id in gene_ids_pert:
        match = np.where(gene_ids_cxg == g_id)[0]
        if len(match) > 0:
            indices_map.append(match[0])
        else:
            indices_map.append(np.nan)
    indices_map = np.array(indices_map)

    gene_ids_pert_samples = f["obs/__categories/gene_id"][:]

    sample_filter = []
    for g_id in gene_ids_pert_samples:
        if (g_id not in gene_ids_pert) and (g_id != b"non-targeting"):
            sample_filter.append(np.nan)
        elif (g_id not in gene_ids_cxg) and (g_id != b"non-targeting"):
            sample_filter.append(np.nan)
        elif g_id != b"non-targeting":
            match = np.where(gene_ids_cxg == g_id)[0]
            sample_filter.append(match[0])
        else:
            sample_filter.append("control")
    sample_filter = np.array(sample_filter, dtype="object")

    matrix = f["X"][:].astype("int32")
    new_matrix = np.zeros((310385, 19331), dtype="int32")
    for ix, l in tqdm(enumerate(indices_map)):
        if ~np.isnan(l):
            new_matrix[:, l.astype(int)] = matrix[:, ix]

    keep_sample_indices = np.array(
        [ix for ix, i in enumerate(sample_filter[:-1]) if ~np.isnan(i)]
        + [len(sample_filter) - 1]
    )
    gene_id_indices = f["obs/gene_id"][:]
    new_matrix = new_matrix[np.isin(gene_id_indices, keep_sample_indices)]

    new_pert_array = []
    for ff in f["obs/gene_id"][:]:
        if sample_filter[ff] == "control":
            new_pert_array.append(np.nan)
        elif np.isnan(sample_filter[ff]):
            continue
        else:
            new_pert_array.append(sample_filter[ff])

    new_pert_array = np.array(new_pert_array, allow_pickle=True)

    splits = np.load(files("bento_sc.utils.data").joinpath("split_pert.npz"))
    split = splits["split"]
    matched_control = splits["matched_control"]
    train_control_indices = splits["train_control_indices"]

    f = h5torch.File(args.output_file_path, "w")
    f.register(
        csr_matrix(new_matrix),
        axis="central",
        mode="csr",
        dtype_save="float32",
        dtype_load="float32",
        csr_load_sparse=True,
    )

    f.register(
        train_control_indices,
        axis="unstructured",
        name="train_control_indices",
        dtype_save="int64",
        dtype_load="int64",
    )

    f.register(
        matched_control,
        axis=0,
        name="matched_control",
        dtype_save="int64",
        dtype_load="int64",
    )

    f.register(split, axis=0, name="split", dtype_save="bytes", dtype_load="str")

    f.register(
        new_pert_array,
        axis=0,
        name="perturbed_gene",
        dtype_save="float64",
        dtype_load="float64",
    )

    f.register(
        f_cxg["1/var"][:],
        axis=1,
        name="var",
        dtype_save="bytes",
        dtype_load="str",
    )

    f.close()
    return None


def get_citeseq(args):
    urllib.request.urlretrieve(
        "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE194nnn/GSE194122/suppl/GSE194122%5Fopenproblems%5Fneurips2021%5Fcite%5FBMMC%5Fprocessed%2Eh5ad%2Egz",
        os.path.join(args.base_folder, "./citeseq.h5ad.gz"),
    )

    with gzip.open(os.path.join(args.base_folder, "./citeseq.h5ad.gz"), "rb") as f_in:
        with open(os.path.join(args.base_folder, "./citeseq.h5ad"), "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)

    f = h5py.File(os.path.join(args.base_folder, "./citeseq.h5ad"))
    f.visititems(print)

    central = f["layers/counts"]
    matrix = csr_matrix(
        (central["data"][:], central["indices"][:], central["indptr"][:])
    ).toarray()

    GEX_columns = (
        f["var/__categories/feature_types"][:][f["var/feature_types"][:]] == b"GEX"
    )
    ADT_columns = (
        f["var/__categories/feature_types"][:][f["var/feature_types"][:]] == b"ADT"
    )

    GEX_matrix = matrix[:, np.where(GEX_columns)[0]]
    ADT_matrix = matrix[:, np.where(ADT_columns)[0]]

    donor_id = f["obs/DonorID"][:]
    site = f["obs/Site"][:]
    split = f["obs/__categories/is_train"][:][f["obs/is_train"][:]]
    split[split == b"iid_holdout"] = b"val"
    celltypes = f["obs/cell_type"][:]
    celltype_cats = f["obs/__categories/cell_type"][:]

    gene_ids = f["var/__categories/gene_id"][:][f["var/gene_id"][:]]
    gene_ids_gex = gene_ids[GEX_columns]
    gene_ids_adt = gene_ids[ADT_columns]

    f_cxg = h5py.File(args.sctab_h5t)
    gene_ids_cxg = f_cxg["1/var"][:, 0]

    indices_map = []
    for g_id in gene_ids_gex:
        match = np.where(gene_ids_cxg == g_id)[0]
        if len(match) > 0:
            indices_map.append(match[0])
        else:
            indices_map.append(np.nan)
    indices_map = np.array(indices_map)

    new_matrix = np.zeros((matrix.shape[0], 19331), dtype="int32")
    for ix, l in tqdm(enumerate(indices_map)):
        if ~np.isnan(l):
            new_matrix[:, l.astype(int)] = GEX_matrix[:, ix]

    f = h5torch.File(args.output_file_path, "w")

    f.register(
        csr_matrix(new_matrix),
        axis="central",
        mode="csr",
        dtype_save="float32",
        dtype_load="float32",
        csr_load_sparse=True,
    )

    f.register(
        split.astype(bytes), axis=0, name="split", dtype_save="bytes", dtype_load="str"
    )

    f.register(
        ADT_matrix, axis=0, name="ADT", dtype_save="float32", dtype_load="float32"
    )

    obs = np.stack([donor_id, site, celltypes]).T
    f.register(obs, axis=0, name="obs", dtype_save="int64", dtype_load="int64")

    f.register(
        np.array(["donor_id", "site", "celltypes"]).astype(bytes),
        axis="unstructured",
        name="obs_columbs",
        dtype_save="bytes",
        dtype_load="str",
    )

    f.register(
        celltype_cats.astype(bytes),
        axis="unstructured",
        name="celltype_names",
        dtype_save="bytes",
        dtype_load="str",
    )

    f.register(
        f_cxg["1/var"][:],
        axis=1,
        name="var",
        dtype_save="bytes",
        dtype_load="str",
    )

    f.register(
        gene_ids_adt.astype(bytes),
        axis="unstructured",
        name="gene_ids_adt",
        dtype_save="bytes",
        dtype_load="str",
    )

    f.close()
    return None


def get_el(args):
    urllib.request.urlretrieve(
        "https://datasets.cellxgene.cziscience.com/6d0d9779-f66b-4297-9bbd-4f232651fb99.h5ad",
        os.path.join(args.base_folder, "./embryonic_limb.h5ad"),
    )

    f = h5py.File(os.path.join(args.base_folder, "./embryonic_limb.h5ad"))

    assay_cats = f["obs/assay/categories"][:]
    assay_codes = f["obs/assay/codes"][:]

    cell_type_cats = f["obs/cell_type/categories"][:]
    cell_type_codes = f["obs/cell_type/codes"][:]

    donor_cats = f["obs/donor_id/categories"][:]
    donor_codes = f["obs/donor_id/codes"][:]

    matrix = csr_matrix(
        (f["raw/X/data"][:], f["raw/X/indices"][:], f["raw/X/indptr"][:])
    ).toarray()

    gene_ids = f["var/feature_name/categories"][:][f["var/feature_name/codes"][:]]

    f_cxg = h5py.File(args.sctab_h5t)
    gene_ids_cxg = f_cxg["1/var"][:, 1]

    indices_map = []
    for g_id in gene_ids:
        match = np.where(gene_ids_cxg == g_id)[0]
        if len(match) > 0:
            indices_map.append(match[0])
        else:
            indices_map.append(np.nan)
    indices_map = np.array(indices_map)

    new_matrix = np.zeros((matrix.shape[0], 19331), dtype="int32")
    for ix, l in tqdm(enumerate(indices_map)):
        if ~np.isnan(l):
            new_matrix[:, l.astype(int)] = matrix[:, ix]

    f = h5torch.File(args.output_file_path, "w")

    f.register(
        csr_matrix(new_matrix),
        axis="central",
        mode="csr",
        dtype_save="float32",
        dtype_load="float32",
        csr_load_sparse=True,
    )

    obs = np.stack([assay_codes, donor_codes, cell_type_codes]).T
    f.register(obs, axis=0, name="obs", dtype_save="int64", dtype_load="int64")

    f.register(
        assay_cats.astype(bytes),
        axis="unstructured",
        name="0_assay_categories",
        dtype_save="bytes",
        dtype_load="str",
    )

    f.register(
        donor_cats.astype(bytes),
        axis="unstructured",
        name="1_donor_categories",
        dtype_save="bytes",
        dtype_load="str",
    )

    f.register(
        cell_type_cats.astype(bytes),
        axis="unstructured",
        name="3_cell_type_categories",
        dtype_save="bytes",
        dtype_load="str",
    )

    split = np.full((len(obs)), "test")
    f.register(split, axis=0, name="split", dtype_save="bytes", dtype_load="str")

    f.register(
        f_cxg["1/var"][:],
        axis=1,
        name="var",
        dtype_save="bytes",
        dtype_load="str",
    )

    f.close()
    return None


def get_ga(args):
    urllib.request.urlretrieve(
        "https://datasets.cellxgene.cziscience.com/b9103b10-e021-4c20-9714-60b32928179e.h5ad",
        os.path.join(args.base_folder, "./great_apes.h5ad"),
    )

    f = h5py.File(os.path.join(args.base_folder, "./great_apes.h5ad"))

    assay_cats = f["obs/assay/categories"][:]
    assay_codes = f["obs/assay/codes"][:]

    cell_type_cats = f["obs/cell_type/categories"][:]
    cell_type_codes = f["obs/cell_type/codes"][:]

    dev_cats = f["obs/development_stage/categories"][:]
    dev_codes = f["obs/development_stage/codes"][:]

    donor_cats = f["obs/donor_id/categories"][:]
    donor_codes = f["obs/donor_id/codes"][:]

    matrix = csr_matrix(
        (f["raw/X/data"][:], f["raw/X/indices"][:], f["raw/X/indptr"][:])
    ).toarray()

    gene_ids = f["var/gene"][:]

    f_cxg = h5py.File(args.sctab_h5t)
    gene_ids_cxg = f_cxg["1/var"][:, 0]

    indices_map = []
    for g_id in gene_ids:
        match = np.where(gene_ids_cxg == g_id)[0]
        if len(match) > 0:
            indices_map.append(match[0])
        else:
            indices_map.append(np.nan)
    indices_map = np.array(indices_map)

    new_matrix = np.zeros((matrix.shape[0], 19331), dtype="int32")
    for ix, l in tqdm(enumerate(indices_map)):
        if ~np.isnan(l):
            new_matrix[:, l.astype(int)] = matrix[:, ix]

    f = h5torch.File(args.output_file_path, "w")

    f.register(
        csr_matrix(new_matrix),
        axis="central",
        mode="csr",
        dtype_save="float32",
        dtype_load="float32",
        csr_load_sparse=True,
    )

    obs = np.stack([assay_codes, dev_codes, donor_codes, cell_type_codes]).T
    f.register(obs, axis=0, name="obs", dtype_save="int64", dtype_load="int64")

    f.register(
        assay_cats.astype(bytes),
        axis="unstructured",
        name="0_assay_categories",
        dtype_save="bytes",
        dtype_load="str",
    )

    f.register(
        dev_cats.astype(bytes),
        axis="unstructured",
        name="1_development_stage_categories",
        dtype_save="bytes",
        dtype_load="str",
    )

    f.register(
        donor_cats.astype(bytes),
        axis="unstructured",
        name="2_donor_categories",
        dtype_save="bytes",
        dtype_load="str",
    )

    f.register(
        cell_type_cats.astype(bytes),
        axis="unstructured",
        name="3_cell_type_categories",
        dtype_save="bytes",
        dtype_load="str",
    )

    split = np.full((len(obs)), "test")
    f.register(split, axis=0, name="split", dtype_save="bytes", dtype_load="str")

    f.register(
        f_cxg["1/var"][:],
        axis=1,
        name="var",
        dtype_save="bytes",
        dtype_load="str",
    )

    f.close()
    return None


def get_ci(args):
    urllib.request.urlretrieve(
        "https://datasets.cellxgene.cziscience.com/e1d1d48b-f0ac-41e6-b7d7-283f709b6421.h5ad",
        os.path.join(args.base_folder, "./circ_imm.h5ad"),
    )

    f = h5py.File(os.path.join(args.base_folder, "./circ_imm.h5ad"))

    assay_cats = f["obs/assay/categories"][:]
    assay_codes = f["obs/assay/codes"][:]

    cell_type_cats = f["obs/cell_type/categories"][:]
    cell_type_codes = f["obs/cell_type/codes"][:]

    donor_cats = f["obs/donor_id/categories"][:]
    donor_codes = f["obs/donor_id/codes"][:]

    matrix = csr_matrix(
        (f["raw/X/data"][:], f["raw/X/indices"][:], f["raw/X/indptr"][:])
    ).toarray()

    gene_ids = f["var/feature_name/categories"][:][f["var/feature_name/codes"][:]]

    f_cxg = h5py.File(args.sctab_h5t)
    gene_ids_cxg = f_cxg["1/var"][:, 1]

    indices_map = []
    for g_id in gene_ids:
        match = np.where(gene_ids_cxg == g_id)[0]
        if len(match) > 0:
            indices_map.append(match[0])
        else:
            indices_map.append(np.nan)
    indices_map = np.array(indices_map)

    new_matrix = np.zeros((matrix.shape[0], 19331), dtype="int32")
    for ix, l in tqdm(enumerate(indices_map)):
        if ~np.isnan(l):
            new_matrix[:, l.astype(int)] = matrix[:, ix]

    f = h5torch.File(args.output_file_path, "w")

    f.register(
        csr_matrix(new_matrix),
        axis="central",
        mode="csr",
        dtype_save="float32",
        dtype_load="float32",
        csr_load_sparse=True,
    )

    obs = np.stack([assay_codes, donor_codes, cell_type_codes]).T
    f.register(obs, axis=0, name="obs", dtype_save="int64", dtype_load="int64")

    f.register(
        assay_cats.astype(bytes),
        axis="unstructured",
        name="0_assay_categories",
        dtype_save="bytes",
        dtype_load="str",
    )

    f.register(
        donor_cats.astype(bytes),
        axis="unstructured",
        name="1_donor_categories",
        dtype_save="bytes",
        dtype_load="str",
    )

    f.register(
        cell_type_cats.astype(bytes),
        axis="unstructured",
        name="3_cell_type_categories",
        dtype_save="bytes",
        dtype_load="str",
    )

    split = np.full((len(obs)), "test")
    f.register(split, axis=0, name="split", dtype_save="bytes", dtype_load="str")

    f.register(
        f_cxg["1/var"][:],
        axis=1,
        name="var",
        dtype_save="bytes",
        dtype_load="str",
    )

    f.close()
    return None


if __name__ == "__main__":
    main()



import os

os.environ["OMP_NUM_THREADS"] = "4"  # export OMP_NUM_THREADS=1
os.environ["OPENBLAS_NUM_THREADS"] = "4"  # export OPENBLAS_NUM_THREADS=1
os.environ["MKL_NUM_THREADS"] = "4"  # export MKL_NUM_THREADS=1
os.environ["VECLIB_MAXIMUM_THREADS"] = "4"  # export VECLIB_MAXIMUM_THREADS=1
os.environ["NUMEXPR_NUM_THREADS"] = "4"  # export NUMEXPR_NUM_THREADS=1
os.environ["PYTORCH_CUDA_ALLOC_CONF"] = "max_split_size_mb:2048"
import torch
import numpy as np
from bento_sc.data import BentoDataModule
from bento_sc.models import BentoTransformer
from bento_sc.utils.config import Config
import argparse
from scipy.sparse import csr_matrix
import pandas as pd
from sklearn.decomposition import PCA
import h5torch
import anndata as ad
import bbknn
import scib
import scanpy as sc
import sys


class CustomFormatter(
    argparse.ArgumentDefaultsHelpFormatter, argparse.MetavarTypeHelpFormatter
):
    pass


def main():
    parser = argparse.ArgumentParser(
        description="Script for batch correction evaluation.",
        formatter_class=CustomFormatter,
    )

    parser.add_argument(
        "embed_or_correct",
        type=str,
        metavar="embed_or_correct",
        choices=["embed", "correct"],
        help="This script contains two subroutines: (1) creating embeddings from a pre-trained model, and (2) performing batch correction and computing scIB scores.",
    )

    args = parser.parse_args(sys.argv[1:2])

    if args.embed_or_correct == "embed":
        embed_parser = get_embed_parser()
        args = embed_parser.parse_args(sys.argv[2:])
        embed(args)

    if args.embed_or_correct == "correct":
        correct_parser = get_correct_parser()
        args = correct_parser.parse_args(sys.argv[2:])
        correct(args)


def get_embed_parser():
    parser = argparse.ArgumentParser(
        description="Create embeddings from a pre-trained model.",
        formatter_class=CustomFormatter,
    )
    parser.add_argument(
        "config_path", type=str, metavar="config_path", help="config_path"
    )
    parser.add_argument(
        "checkpoint",
        type=str,
        metavar="checkpoint",
        help="pre-trained model checkpoint",
    )
    parser.add_argument("save_path", type=str, metavar="save_path", help="save_path")
    parser.add_argument(
        "--data_path",
        type=str,
        default=None,
        help="Data file. Overrides value in config file if specified",
    )
    return parser


def get_correct_parser():
    parser = argparse.ArgumentParser(
        description="Perform batch correction and compute scIB scores.",
        formatter_class=CustomFormatter,
    )
    parser.add_argument(
        "h5t_data_file", type=str, metavar="h5t_data_file", help="h5torch data file"
    )
    parser.add_argument(
        "input_model_embeds",
        type=str,
        metavar="input_model_embeds",
        help="save_path from previous step",
    )
    parser.add_argument(
        "output_h5ad",
        type=str,
        metavar="output_h5ad",
        help="results will be written in a h5ad file",
    )
    parser.add_argument(
        "batch_col",
        type=str,
        metavar="batch_col",
        help="index of column in 0/obs in h5t file where batch effects are",
    )
    parser.add_argument(
        "ct_col",
        type=str,
        metavar="ct_col",
        help="index of column in 0/obs in h5t file where celltypes are",
    )
    return parser


def embed(args):
    config = Config(args.config_path)

    if args.data_path is not None:
        config["data_path"] = args.data_path

    dm = BentoDataModule(config)
    dm.setup(None)

    model = BentoTransformer.load_from_checkpoint(args.checkpoint)

    device_ = "cuda:%s" % config.devices[0]
    model = model.to(device_).to(torch.bfloat16).eval()

    obs = []
    embeds = []
    with torch.no_grad():
        for batch in dm.predict_dataloader():
            batch["gene_index"] = batch["gene_index"].to(model.device)
            batch["gene_counts"] = batch["gene_counts"].to(model.device)
            batch["gene_counts_true"] = batch["gene_counts_true"].to(model.device)

            if not model.config.discrete_input:
                batch["gene_counts"] = batch["gene_counts"].to(model.dtype)
            else:
                batch["gene_counts"] = batch["gene_counts"].float()

            y = model(batch)

            embeds.append(y[:, 0].cpu())
            obs.append(batch["0/obs"])

    embeds = torch.cat(embeds).float().numpy()
    obs = torch.cat(obs).numpy()

    np.savez(args.save_path, obs=obs, embeds=embeds)
    return None


def correct(args):
    f = h5torch.File(args.h5t_data_file)
    f = f.to_dict()

    matrix = csr_matrix(
        (f["central/data"][:], f["central/indices"][:], f["central/indptr"][:]),
        shape=(f["0/obs"].shape[0], f["1/var"].shape[0]),
    )

    adata = ad.AnnData(matrix)
    adata.obs = pd.DataFrame(
        f["0/obs"], columns=np.arange(f["0/obs"].shape[1]).astype(str)
    )
    adata.var = pd.DataFrame(
        f["1/var"], columns=np.arange(f["1/var"].shape[1]).astype(str)
    )

    adata.obs[args.batch_col] = adata.obs[args.batch_col].astype("category")
    adata.obs[args.ct_col] = adata.obs[args.ct_col].astype("category")

    file = np.load(args.input_model_embeds)
    adata.obsm["X_emb"] = file["embeds"]

    embeds_pca = PCA(n_components=50).fit_transform(adata.obsm["X_emb"])

    adata.obsm["X_pca"] = embeds_pca

    bbknn.bbknn(adata, batch_key=args.batch_col)

    sc.tl.umap(adata)

    clisi, ilisi = scib.me.lisi_graph(
        adata, batch_key=args.batch_col, label_key=args.ct_col, type_="knn", n_cores=16
    )
    graph_conn = scib.me.graph_connectivity(adata, label_key=args.ct_col)

    scib.cl.cluster_optimal_resolution(
        adata, cluster_key="iso_label", label_key=args.ct_col
    )
    iso_f1 = scib.me.isolated_labels_f1(
        adata, batch_key=args.batch_col, label_key=args.ct_col, embed=None
    )
    ari = scib.me.ari(adata, cluster_key="iso_label", label_key=args.ct_col)
    nmi = scib.me.nmi(adata, cluster_key="iso_label", label_key=args.ct_col)

    adata.uns["scores"] = {
        "iLISI": ilisi,
        "Graph Connectivity": graph_conn,
        "cLISI": clisi,
        "ARI": ari,
        "NMI": nmi,
        "Isolated F1": iso_f1,
    }
    adata.write(args.output_h5ad)
    return None


if __name__ == "__main__":
    main()



from bento_sc.utils.config import Config
from bento_sc.data import BentoDataModule
from bento_sc.models import BentoTransformer
from bento_sc.utils.metrics import pearson_batch_masked
import torch
import numpy as np
import argparse


def boolean(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


def main():
    class CustomFormatter(
        argparse.ArgumentDefaultsHelpFormatter, argparse.MetavarTypeHelpFormatter
    ):
        pass

    parser = argparse.ArgumentParser(
        description="Script for gene expression upscaling evaluation.",
        formatter_class=CustomFormatter,
    )
    parser.add_argument(
        "config_path", type=str, metavar="config_path", help="config_path"
    )
    parser.add_argument(
        "checkpoint", type=str, metavar="checkpoint", help="checkpoint path"
    )
    parser.add_argument(
        "--data_path",
        type=str,
        default=None,
        help="Data file. Overrides value in config file if specified",
    )
    parser.add_argument(
        "--clf_output",
        type=boolean,
        default=True,
        help="Set this flag to true if the model produced count predictions through multi-class clf.",
    )
    parser.add_argument(
        "--model_preds_logp1",
        type=boolean,
        default=False,
        help="Set this flag to true if the pre-trained model predicted log transformed counts.",
    )

    args = parser.parse_args()

    config = Config(args.config_path)
    if args.data_path is not None:
        config["data_path"] = args.data_path

    dm = BentoDataModule(config)
    dm.setup(None)

    device_ = "cuda:%s" % config.devices[0]

    model = BentoTransformer.load_from_checkpoint(args.checkpoint)
    model = model.to(device_).to(torch.bfloat16).eval()

    counts = []
    trues = []

    with torch.no_grad():
        for batch in dm.predict_dataloader():
            batch["gene_index"] = batch["gene_index"].to(model.device)
            batch["gene_counts"] = batch["gene_counts"].to(model.device)
            batch["gene_counts_true"] = batch["gene_counts_true"].to(model.device)

            if not model.config.discrete_input:
                batch["gene_counts"] = batch["gene_counts"].to(model.dtype)
            else:
                batch["gene_counts"] = batch["gene_counts"].float()

            y = model(batch)

            libsizes = (
                batch["gene_counts_true"].sum(1)
                + (batch["gene_counts_true"] == -1).sum(1)
            )[:, None]

            count_predictions = model.loss.predict(
                y[:, 1:],
                gene_ids=batch["gene_index"],
                libsize=libsizes,
            )
            if isinstance(count_predictions, tuple):
                count_predictions = count_predictions[0]

            counts.append(count_predictions.cpu())
            trues.append(batch["gene_counts_true"].cpu())

    if args.clf_output:
        for temp in [0.01, 0.05, 0.1, 0.5, 1, 5, 10, 50, 100]:
            pearsons = []
            for batch_ix in range(len(counts)):

                multiplier = torch.arange(counts[batch_ix].shape[-1])[None, None, :]

                predicted_as_count = (
                    torch.nn.functional.softmax(counts[batch_ix].float() * temp, -1)
                    * multiplier
                ).sum(-1)

                true_count = trues[batch_ix].float()

                pearsons.append(
                    pearson_batch_masked(predicted_as_count, true_count).numpy()
                )

            print(temp, np.concatenate(pearsons).mean())
    else:
        pearsons = []
        for batch_ix in range(len(counts)):
            if args.model_preds_logp1:
                predicted_as_count = torch.expm1(counts[batch_ix].float())
            else:
                predicted_as_count = counts[batch_ix].float()
            true_count = trues[batch_ix].float()
            pearsons.append(
                pearson_batch_masked(predicted_as_count, true_count).numpy()
            )
        print(np.concatenate(pearsons).mean())


if __name__ == "__main__":
    main()



from bento_sc.data import BentoDataModule
from bento_sc.utils.config import Config
from bento_sc.models import CLSTaskTransformer, BentoTransformer
from lightning.pytorch.callbacks import ModelCheckpoint, EarlyStopping
from lightning.pytorch.loggers import TensorBoardLogger
from lightning.pytorch.plugins.environments import LightningEnvironment
from lightning.pytorch import Trainer
import argparse


def boolean(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


def main():
    class CustomFormatter(
        argparse.ArgumentDefaultsHelpFormatter, argparse.MetavarTypeHelpFormatter
    ):
        pass

    parser = argparse.ArgumentParser(
        description="Fine-tuning script for cell-type identification evaluation.",
        formatter_class=CustomFormatter,
    )

    parser.add_argument(
        "config_path", type=str, metavar="config_path", help="config_path"
    )
    parser.add_argument(
        "checkpoint", type=str, metavar="checkpoint", help="checkpoint path"
    )
    parser.add_argument("logs_path", type=str, metavar="logs_path", help="logs_path")
    parser.add_argument(
        "--data_path",
        type=str,
        default=None,
        help="Data file. Overrides value in config file if specified",
    )
    parser.add_argument(
        "--lr",
        type=float,
        default=None,
        help="Learning rate. Overrides value in config file if specified",
    )
    parser.add_argument(
        "--batch_size",
        type=int,
        default=None,
        help="Batch size. Overrides value in config file if specified",
    )
    parser.add_argument(
        "--n_workers",
        type=int,
        default=None,
        help="Num workers. Overrides value in config file if specified",
    )
    parser.add_argument(
        "--prefetch_factor",
        type=int,
        default=None,
        help="Prefetch Factor of dataloader. Overrides value in config file if specified",
    )
    parser.add_argument(
        "--tune_mode",
        type=boolean,
        default=False,
        help="Don't pre-train whole model but run small experiment.",
    )

    args = parser.parse_args()

    config = Config(args.config_path)

    if args.lr is not None:
        config["lr"] = args.lr
    if args.data_path is not None:
        config["data_path"] = args.data_path
    if args.batch_size is not None:
        config["batch_size"] = args.batch_size
    if args.n_workers is not None:
        config["n_workers"] = args.n_workers
    if args.prefetch_factor is not None:
        config["prefetch_factor"] = args.prefetch_factor

    dm = BentoDataModule(config)

    dm.setup(None)

    model = CLSTaskTransformer(config)
    pretrained_model = BentoTransformer.load_from_checkpoint(args.checkpoint)

    pretrained_dict = pretrained_model.state_dict()
    model_dict = model.state_dict()
    pretrained_dict_new = {
        k: v
        for k, v in pretrained_dict.items()
        if not k.startswith(("nce_loss", "ct_clf_loss", "loss"))
    }

    model_dict.update(pretrained_dict_new)
    model.load_state_dict(model_dict)

    val_ckpt = ModelCheckpoint(monitor="val_macroacc", mode="max")
    callbacks = [
        val_ckpt,
        EarlyStopping(monitor="val_macroacc", patience=40, mode="max"),
    ]

    logger = TensorBoardLogger(
        "/".join(args.logs_path.split("/")[:-1]),
        name=args.logs_path.split("/")[-1],
    )

    if args.tune_mode:
        max_steps = int(50_000 // (config.batch_size / 128)) + 1
        val_check_interval = int(1_000 // (config.batch_size / 128) ** 0.5)
    else:
        max_steps = int(500_000 // (config.batch_size / 128)) + 1
        val_check_interval = int(5_000 // (config.batch_size / 128) ** 0.5)

    trainer = Trainer(
        accelerator="gpu",
        devices=config.devices,
        strategy="auto",
        plugins=[LightningEnvironment()],
        max_steps=max_steps,
        val_check_interval=val_check_interval,
        check_val_every_n_epoch=None,
        gradient_clip_val=1,
        callbacks=callbacks,
        logger=logger,
        precision="bf16-true",
        use_distributed_sampler=(True if config.return_zeros else False),
    )

    trainer.fit(model, dm.train_dataloader(), dm.val_dataloader())

    res = trainer.validate(model, dm.val_dataloader(), ckpt_path="best")
    print(res)
    res = trainer.validate(model, dm.test_dataloader(), ckpt_path="best")
    print(res)


if __name__ == "__main__":
    main()



from bento_sc.data import CellSampleProcessor, SequentialPreprocessor, BatchCollater
import h5torch
from tqdm import tqdm
import torch


train = h5torch.Dataset(
    "../data/circ_imm.h5t",
    sample_processor=CellSampleProcessor(
        SequentialPreprocessor(*[]), return_zeros=True
    ),
    in_memory=False,
    subset=("0/split", "test"),
)

dataloader = torch.utils.data.DataLoader(
    train,
    num_workers=4,
    batch_size=512,
    shuffle=False,
    pin_memory=True,
    collate_fn=BatchCollater(),
)

dict_keeper = []
for i in range(19331):
    dict_keeper.append(dict())


for batch in tqdm(dataloader):
    for i in range(19331):
        s, c = torch.unique(batch["gene_counts"][:, i], return_counts=True)
        for ss, cc in zip(s, c):
            if ss.item() not in dict_keeper[i]:
                dict_keeper[i][ss.item()] = cc.item()
            else:
                dict_keeper[i][ss.item()] += cc.item()


def mu(gene_c):
    total = 0
    running_sum = 0
    for k, v in gene_c.items():
        running_sum += k * v
        total += v

    return running_sum / total


def var(gene_c, mu):
    total = 0
    running_sum = 0
    for k, v in gene_c.items():
        running_sum += ((k - mu) ** 2) * v
        total += v

    return running_sum / total


mus = []
vars = []
for i in range(19331):
    gene_c = dict_keeper[i]
    mu_i = mu(gene_c)
    var_i = var(gene_c, mu_i)
    mus.append(mu_i)
    vars.append(var_i)

import matplotlib.pyplot as plt
import numpy as np
from skmisc.loess import loess

not_const = np.array(vars) > 0
x = np.log10(np.array(mus))[not_const]
y = np.log10(np.array(vars))[not_const]
model = loess(x, y, span=0.3, degree=2)
model.fit()


def safe_var(gene_c, mu, reg_var):
    total = sum(list(gene_c.values()))
    running_sum = 0
    for k, v in gene_c.items():
        val = (k - mu) / reg_var
        running_sum += (np.minimum(val, total**0.5) ** 2) * v

    return running_sum / (total - 1)


final_var = np.zeros(19331)
for ix, i in enumerate(np.where(not_const)[0]):
    gene_c = dict_keeper[i]
    mu_i = mus[i]
    var_i = model.outputs.fitted_values[ix]
    final_var[i] = safe_var(gene_c, mu_i, var_i)

np.save("./hvg.npy", final_var)



import torch
import torch.nn as nn
import torch.nn.functional as F


class BentoLoss(nn.Module):
    def __init__(self, in_dim, out_dim, reduction="mean"):
        super().__init__()

        self.output_head = nn.Linear(in_dim, out_dim)

        if reduction == "mean":
            self.reduce = torch.mean
        elif reduction == "sum":
            self.reduce = torch.sum
        elif reduction == "none":
            self.reduce = lambda x: x

    def predict(self, *args):
        raise NotImplementedError

    def forward(self, *args):
        raise NotImplementedError

    def loss(self, *args):
        raise NotImplementedError


class BinCE(BentoLoss):
    def __init__(self, dim, n_bins, reduction="mean"):
        super().__init__(dim, n_bins, reduction=reduction)

    def predict(self, x, **kwargs):
        return self.output_head(x)

    def loss(self, inputs, targets):
        return self.reduce(F.cross_entropy(inputs, targets, reduction="none"))

    def forward(self, x, targets, train_on=None, **kwargs):
        if train_on is not None:
            y = self.predict(x)[train_on]
            return self.loss(y, targets[train_on])
        else:
            y = self.predict(x)
            return self.loss(y, targets)


class CountMSE(BentoLoss):
    def __init__(
        self, dim, exp_output=True, lib_norm=False, reduction="mean", plus_one=False
    ):
        super().__init__(dim, 1, reduction=reduction)
        assert not (not exp_output and lib_norm), "lib norm true needs exp output True"
        self.exp_output = exp_output
        self.lib_norm = lib_norm
        self.plus_one = plus_one

    def predict(self, x, libsize=None, **kwargs):
        y = self.output_head(x).squeeze(-1)
        if not self.exp_output:
            return y
        if not self.plus_one:
            if self.lib_norm:
                return F.softmax(y, -1) * libsize
            else:
                return torch.clamp(y.exp(), max=1e7)
        else:
            if self.lib_norm:
                return F.softmax(y, -1) * (libsize - y.shape[-1]) + 1
            else:
                return torch.clamp(y.exp() + 1, max=1e7)

    def loss(self, inputs, targets):
        return self.reduce(F.mse_loss(inputs, targets, reduction="none"))

    def forward(self, x, targets, train_on=None, **kwargs):
        y = self.predict(x, libsize=targets.sum(1)[:, None])
        if train_on is not None:
            return self.loss(y[train_on], targets.to(y.dtype)[train_on])
        else:
            return self.loss(y, targets.to(y.dtype))


class PoissonNLL(BentoLoss):
    def __init__(
        self,
        dim,
        lib_norm=False,
        reduction="mean",
        omit_last_term=True,
        zero_truncated=False,
    ):
        super().__init__(dim, 1, reduction=reduction)
        self.omit = omit_last_term
        self.lib_norm = lib_norm
        self.zero_trunc = zero_truncated

    def predict(self, x, libsize=None, **kwargs):
        y = self.output_head(x).squeeze(-1)
        if self.zero_trunc:
            if self.lib_norm:
                return F.softmax(y, -1) * (libsize - y.shape[-1]) + 1
            else:
                return torch.clamp(y.exp() + 1, max=1e7)
        else:
            if self.lib_norm:
                return F.softmax(y, -1) * libsize
            else:
                return torch.clamp(y.exp(), max=1e7)

    def loss(self, inputs, targets):
        if self.zero_trunc:
            stabilized_term = torch.where(
                inputs < 10,
                (torch.clamp(inputs, max=10).exp() - 1 + 1e-8).log(),
                inputs,
            )
            loss = stabilized_term - targets * (inputs + 1e-8).log()
        else:
            loss = inputs - targets * (inputs + 1e-8).log()

        if self.omit:
            return self.reduce(loss)
        else:
            return self.reduce(loss + torch.lgamma(targets + 1))

    def forward(self, x, targets, train_on=None, **kwargs):
        y = self.predict(x, libsize=targets.sum(1)[:, None])
        if train_on is not None:
            return self.loss(y[train_on], targets[train_on])
        else:
            return self.loss(y, targets)


class NegativeBinomialNLL(BentoLoss):
    def __init__(
        self,
        dim,
        lib_norm=False,
        n_genes=19331,
        fixed_dispersion=False,
        reduction="mean",
        omit_last_term=True,
        zero_truncated=False,
    ):
        super().__init__(dim, (1 if fixed_dispersion else 2), reduction=reduction)

        self.omit = omit_last_term
        self.fixed_dispersion = fixed_dispersion
        if self.fixed_dispersion:
            self.dispersions = nn.Embedding(n_genes, 1)
        self.lib_norm = lib_norm
        self.zero_trunc = zero_truncated

    def predict(self, x, gene_ids=None, libsize=None, **kwargs):
        y = self.output_head(x)
        if self.fixed_dispersion:
            mus = y.squeeze(-1)
            log_thetas = self.dispersions(gene_ids).squeeze(-1)
        else:
            mus, log_thetas = y[..., 0], y[..., 1]

        if self.zero_trunc:
            if self.lib_norm:
                mus = F.softmax(mus, -1) * (libsize - mus.shape[-1]) + 1
            else:
                mus = torch.clamp(mus.exp() + 1, max=1e7)
        else:
            if self.lib_norm:
                mus = F.softmax(mus, -1) * libsize
            else:
                mus = torch.clamp(mus.exp(), max=1e7)

        return mus, log_thetas

    def loss(self, mus, log_thetas, targets):

        eps = torch.finfo(mus.dtype).tiny

        log_thetas += eps
        thetas = log_thetas.exp()
        mus += eps

        loss = (
            torch.lgamma(thetas.float())
            - torch.lgamma((targets + thetas).float())
            + targets * (thetas + mus).log()
            - thetas * log_thetas
            - targets * mus.log()
        ).to(mus.dtype)
        if self.zero_trunc:
            stabilized_term = torch.where(
                torch.logical_or(
                    (mus - 1 + 1e-8) * (thetas + 1e-8) > 15, mus + thetas > 15
                ),
                thetas * (thetas + mus).log(),
                (
                    torch.clamp(thetas + mus, max=15) ** torch.clamp(thetas, max=15)
                    - torch.clamp(thetas, max=15) ** torch.clamp(thetas, max=15)
                    + 1e-8
                ).log(),
            )
            loss += stabilized_term
        else:
            loss += thetas * (thetas + mus).log()

        if self.omit:
            return self.reduce(loss)
        else:
            return self.reduce(loss + torch.lgamma(targets + 1))

    def forward(self, x, targets, gene_ids=None, train_on=None):
        mus, log_thetas = self.predict(
            x, gene_ids=gene_ids, libsize=targets.sum(1)[:, None]
        )

        if train_on is not None:
            return self.loss(mus[train_on], log_thetas[train_on], targets[train_on])
        else:
            return self.loss(mus, log_thetas, targets)


class ZeroInflatedNegativeBinomialNLL(NegativeBinomialNLL):
    def __init__(
        self,
        dim,
        lib_norm=False,
        n_genes=19331,
        fixed_dispersion=False,
        reduction="mean",
        omit_last_term=True,
    ):
        super().__init__(
            dim,
            lib_norm=lib_norm,
            n_genes=n_genes,
            fixed_dispersion=fixed_dispersion,
            reduction=reduction,
            omit_last_term=omit_last_term,
        )

        self.out_pi = nn.Linear(dim, 1)

    def predict(self, x, gene_ids=None, libsize=None, **kwargs):
        mus, log_thetas = super().predict(x, gene_ids=gene_ids, libsize=libsize)
        pis = self.out_pi(x).squeeze(-1)
        return mus, log_thetas, pis

    def loss(self, mus, log_thetas, pis, targets):
        NB_NLL_loss = super().loss(mus, log_thetas, targets)

        eps = torch.finfo(mus.dtype).tiny
        mus += eps
        log_thetas += eps
        thetas = log_thetas.exp()

        indices = targets > 0

        NLLifzero = F.softplus(-pis) - F.softplus(
            -pis + thetas * (log_thetas - (thetas + mus).log())
        )
        NLLifnotzero = pis + F.softplus(-pis) + NB_NLL_loss

        return self.reduce(NLLifnotzero * indices + NLLifzero * ~indices)

    def forward(self, x, targets, gene_ids=None):
        mus, log_thetas, pis = self.predict(x, gene_ids=gene_ids, libsize=targets.sum())
        return self.loss(mus, log_thetas, pis, targets)


class NCELoss(BentoLoss):
    def __init__(
        self,
        dim,
        embed_dim,
        reduction="mean",
        temperature=1,
    ):
        super().__init__(dim, embed_dim, reduction=reduction)
        self.t = temperature

    def predict(self, x, **kwargs):
        return self.output_head(x)

    def loss(self, inputs):
        n = len(inputs)
        targets = torch.arange(n).view(n // 2, 2).fliplr().view(n).to(inputs.device)

        inputs = F.normalize(inputs, dim=-1)
        inputs = (inputs @ inputs.T) / self.t
        inputs.fill_diagonal_(float("-inf"))
        return self.reduce(F.cross_entropy(inputs, targets, reduction="none"))

    def forward(self, x):
        y = self.predict(x)
        return self.loss(y)


class CellTypeClfLoss(BentoLoss):
    def __init__(
        self,
        dim,
        n_classes,
        reduction="mean",
    ):
        super().__init__(dim, n_classes, reduction=reduction)

    def predict(self, x, **kwargs):
        return self.output_head(x)

    def loss(self, inputs, targets):
        return self.reduce(F.cross_entropy(inputs, targets, reduction="none"))

    def forward(self, x, targets):
        y = self.predict(x)
        return self.loss(y, targets)


class ModalityPredictionLoss(BentoLoss):
    def __init__(
        self,
        dim,
        n_classes,
        reduction="mean",
    ):
        super().__init__(dim, n_classes, reduction=reduction)

    def predict(self, x, **kwargs):
        return self.output_head(x)

    def loss(self, inputs, targets):
        return self.reduce(
            F.mse_loss(inputs, torch.log1p(targets.to(inputs.dtype)), reduction="none")
        )

    def forward(self, x, targets):
        y = self.predict(x)
        return self.loss(y, targets)



import h5torch
from tqdm import tqdm
import numpy as np

f = h5torch.File("../cellxgene.h5t")

ss = np.random.choice(f["central/data"].shape[0], size=10_000, replace=True)
ss = np.unique(np.sort(ss))
counts = {}
for c in tqdm(ss):
    counts_section = f["central/data"][c : c + 100_000]
    for s, c in zip(*np.unique(counts_section, return_counts=True)):
        if s in counts:
            counts[s] += c
        else:
            counts[s] = c


counts_inarray = np.zeros((max(list(counts.keys()))).astype(int) + 1)
for k, v in counts.items():
    counts_inarray[k.astype(int)] = v


p = counts_inarray[1:] / counts_inarray.sum()


bins = []
n_bins = 500
rolling_percent = 0
start_loc = len(p) - 1
for i in range(len(p) - 1, -1, -1):
    rolling_percent += p[i]
    if rolling_percent > 1 / n_bins:
        bins.append(start_loc)
        start_loc = i
        rolling_percent = 0

bins = np.array(bins[::-1])
bins[-1] = 1e9
np.savetxt("../bins500.txt", bins)



from bento_sc.utils.config import Config
from bento_sc.data import *
from bento_sc.models import BentoTransformer
import torch
import pandas as pd
import numpy as np
from importlib.resources import files
from tqdm import tqdm
from sklearn.linear_model import LinearRegression
import argparse
import h5py
import os
import sys


def boolean(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


class CustomFormatter(
    argparse.ArgumentDefaultsHelpFormatter, argparse.MetavarTypeHelpFormatter
):
    pass


def main():
    parser = argparse.ArgumentParser(
        description="Script for GRN inference evaluation.",
        formatter_class=CustomFormatter,
    )

    parser.add_argument(
        "embed_or_eval",
        type=str,
        metavar="embed_or_eval",
        choices=["embed", "eval"],
        help="This script contains two subroutines: (1) creating embeddings from a pre-trained model, and (2) inferring a GRN and evaluating it through a regression proxy task.",
    )

    args = parser.parse_args(sys.argv[1:2])

    if args.embed_or_eval == "embed":
        embed_parser = get_embed_parser()
        args = embed_parser.parse_args(sys.argv[2:])
        embed(args)

    if args.embed_or_eval == "eval":
        eval_parser = get_eval_parser()
        args = eval_parser.parse_args(sys.argv[2:])
        eval_grn(args)


def get_embed_parser():
    parser = argparse.ArgumentParser(
        description="Create embeddings from a pre-trained model.",
        formatter_class=CustomFormatter,
    )
    parser.add_argument(
        "config_path", type=str, metavar="config_path", help="config_path"
    )
    parser.add_argument(
        "checkpoint",
        type=str,
        metavar="checkpoint",
        help="pre-trained model checkpoint",
    )
    parser.add_argument("save_path", type=str, metavar="save_path", help="save_path")
    parser.add_argument(
        "--data_path",
        type=str,
        default=None,
        help="Data file. Overrides value in config file if specified",
    )
    parser.add_argument(
        "--counts_as_pos",
        type=boolean,
        default=False,
        help="Flag to enable if your pre-trained models treated counts as positional indices.",
    )

    return parser


def get_eval_parser():
    parser = argparse.ArgumentParser(
        description="Infer a GRN from embeddings and evaluate.",
        formatter_class=CustomFormatter,
    )
    parser.add_argument(
        "config_path", type=str, metavar="config_path", help="config_path"
    )
    parser.add_argument(
        "path_to_embeddings",
        type=str,
        metavar="path_to_embeddings",
        help="path_to_embeddings",
    )
    parser.add_argument(
        "pertdata_path", type=str, help="Path to external validation perturbation data"
    )
    parser.add_argument(
        "scenic_database", type=str, help="SCENIC motif to gene binding database"
    )
    parser.add_argument(
        "--data_path",
        type=str,
        default=None,
        help="Data file. Overrides value in config file if specified",
    )
    parser.add_argument(
        "--test_mode",
        type=str,
        default="val",
        choices=["val", "test"],
        help="val or test",
    )
    parser.add_argument(
        "--n_features",
        type=int,
        default=5,
        help="max number of features to use in the regression task",
    )
    parser.add_argument(
        "--motifs_per_gene",
        type=int,
        default=500,
        help="max number of motifs per gene to allow",
    )
    parser.add_argument(
        "--n_genes_to_train_on",
        type=int,
        default=200,
        help="number of regression models to train.",
    )
    return parser


def embed(args):
    config = Config(args.config_path)
    if args.data_path is not None:
        config["data_path"] = args.data_path

    dm = BentoDataModule(config)
    dm.setup(None)

    model = BentoTransformer.load_from_checkpoint(args.checkpoint)

    model = model.eval().to(config.devices[0]).to(torch.bfloat16)

    embeddings_per_gene = {
        "Myeloid cells": {i: [] for i in range(19331)},
        "B cells": {i: [] for i in range(19331)},
        "NK cells": {i: [] for i in range(19331)},
        "T cells": {i: [] for i in range(19331)},
    }

    key = "gene_index" if not args.counts_as_pos else "gene_counts"

    with torch.no_grad():
        for batch in tqdm(dm.test_dataloader(), total=len(dm.test_dataloader())):
            if not model.config.discrete_input:
                batch["gene_counts"] = (
                    batch["gene_counts"].to(model.dtype).to(model.device)
                )
            else:
                batch["gene_counts"] = batch["gene_counts"].float().to(model.device)

            batch["gene_index"] = batch["gene_index"].to(model.device)
            y = model(batch).cpu()

            for sample in range(len(batch["gene_index"])):
                for gene in range(batch["gene_index"].shape[1]):
                    if batch["gene_counts"][sample, gene] != -1:
                        embeddings_per_gene[batch["0/celltype"][sample]][
                            batch[key][sample, gene].item()
                        ].append(y[:, 1:][sample, gene].cpu())

    embeddings_per_celltype = {}
    for celltype in embeddings_per_gene:
        aggregated_per_cell = {
            k: torch.stack(v).mean(0)
            for k, v in embeddings_per_gene[celltype].items()
            if len(v) > 10
        }

        embeddings = torch.zeros((19331, 512))

        for k1, v1 in aggregated_per_cell.items():
            embeddings[k1] = v1

        embeddings_per_celltype[celltype] = embeddings

    torch.save(
        embeddings_per_celltype["NK cells"],
        os.path.join(args.save_path, "embeddings_NK.pt"),
    )
    torch.save(
        embeddings_per_celltype["T cells"],
        os.path.join(args.save_path, "embeddings_T.pt"),
    )
    torch.save(
        embeddings_per_celltype["Myeloid cells"],
        os.path.join(args.save_path, "embeddings_Myeloid.pt"),
    )
    torch.save(
        embeddings_per_celltype["B cells"],
        os.path.join(args.save_path, "embeddings_B.pt"),
    )
    return None


def eval_grn(args):
    config = Config(args.config_path)
    if args.data_path is not None:
        config["data_path"] = args.data_path

    dm = BentoDataModule(config)
    dm.setup(None)

    embeddings_per_celltype = {
        "NK cells": torch.load(
            os.path.join(args.path_to_embeddings, "embeddings_NK.pt")
        ),
        "T cells": torch.load(os.path.join(args.path_to_embeddings, "embeddings_T.pt")),
        "B cells": torch.load(os.path.join(args.path_to_embeddings, "embeddings_B.pt")),
        "Myeloid cells": torch.load(
            os.path.join(args.path_to_embeddings, "embeddings_Myeloid.pt")
        ),
    }

    gene_lists_celltype = {}
    top_k_celltype = {}
    gene_ids_cxg = dm.train.f["1/var"][:, 1]
    for celltype in embeddings_per_celltype:
        embeddings = embeddings_per_celltype[celltype]

        embeddings_norm = embeddings / (embeddings.norm(dim=1)[:, None] + 1e-8)

        A = torch.einsum("k h, q h -> k q", embeddings_norm, embeddings_norm)

        A_reduced = A[torch.where(A.sum(1))[0]][:, torch.where(A.sum(1))[0]]

        A_reduced = A_reduced - torch.diag(torch.ones(len(A_reduced)))

        top_k = torch.argsort(-(A_reduced).abs(), dim=1)
        gene_list = torch.where(A.sum(1))[0]
        top_k = gene_list[top_k]

        gene_lists_celltype[celltype] = gene_ids_cxg[gene_list.numpy()]
        top_k_celltype[celltype] = gene_ids_cxg[top_k.numpy()]

    motif2gene_db = pd.read_feather(args.scenic_database)
    motif2tf_db = pd.read_table(
        files("bento_sc.utils.data").joinpath(
            "motifs-v10nr_clust-nr.hgnc-m0.001-o0.0.tbl"
        )
    )

    path = files("bento_sc.utils.data").joinpath("allTFs_hg38.txt")
    TFs = np.loadtxt(path, dtype="str").astype(bytes)

    TFs_in_motif2tf = motif2tf_db["gene_name"].values.astype(bytes)

    motif2tf_db = motif2tf_db.iloc[np.isin(TFs_in_motif2tf, TFs)]

    motif2tf_dict = {}
    for m, tf in zip(motif2tf_db["#motif_id"], motif2tf_db["gene_name"]):
        if m not in motif2tf_dict:
            motif2tf_dict[m] = [tf]
        else:
            if tf not in motif2tf_dict[m]:
                motif2tf_dict[m].append(tf)

    normalized = motif2gene_db.iloc[:, :-1].values / (
        motif2gene_db.iloc[:, :-1].values.mean(1)[:, None] + 1e-8
    )
    ranking = np.argsort(normalized, axis=0)[::-1][: args.motifs_per_gene]
    gene_2_tf_dict = {}
    for gene, rank in zip(motif2gene_db.columns[:-1], ranking.T):
        motifs_for_gene = motif2gene_db.iloc[:, -1].values[rank]
        tfs_for_gene = []
        for motif in motifs_for_gene:
            if motif in motif2tf_dict:
                tfs_for_gene += motif2tf_dict[motif]
        gene_2_tf_dict[gene] = list(np.unique(tfs_for_gene))

    gene_2_tf_dict_filt = {k: v for k, v in gene_2_tf_dict.items()}

    f = h5py.File(args.pertdata_path)
    gene_ids_pert = f["var"]["_index"][:]
    gene_ids_scenic = np.array(list(gene_2_tf_dict_filt)).astype(bytes)
    tf_ids_scenic = np.array(
        list((set([t for v in gene_2_tf_dict_filt.values() for t in v])))
    ).astype(bytes)

    possible_gene_ids = np.array(
        list(set(list(gene_ids_pert)).intersection(gene_ids_scenic))
    )

    possible_TF_ids = np.array(
        list(set(list(gene_ids_pert)).intersection(tf_ids_scenic))
    )

    gene_lists_celltype_filt = {}
    top_k_celltype_filt = {}
    for ct in gene_lists_celltype.keys():
        gene_list_ct = gene_lists_celltype[ct]
        top_k_ct = top_k_celltype[ct]

        index_select = np.isin(gene_list_ct, possible_gene_ids)
        gene_lists_celltype_filt[ct] = gene_list_ct[index_select]
        top_k_celltype_filt[ct] = top_k_ct[index_select]

    dataset = f["layers"]["scgen_pearson"][:]

    dataset_f_select = np.isin(gene_ids_pert, possible_gene_ids)
    dataset_genes = dataset[:, dataset_f_select]
    dataset_genes_f = gene_ids_pert[dataset_f_select]

    dataset_f_select = np.isin(gene_ids_pert, possible_TF_ids)
    dataset_tf = dataset[:, dataset_f_select]
    dataset_tf_f = gene_ids_pert[dataset_f_select]

    score_per_celltype = []
    for celltype in gene_lists_celltype:
        index_of_celltype = np.where(
            f["obs"]["cell_type"]["categories"][:].astype(str) == celltype
        )[0][0]
        indices_celltype_in_pertdata = (
            f["obs"]["cell_type"]["codes"][:] == index_of_celltype
        )
        dataset_genes_celltype = dataset_genes[indices_celltype_in_pertdata]
        dataset_tf_celltype = dataset_tf[indices_celltype_in_pertdata]

        groups = f["obs/sm_name/codes"][:][indices_celltype_in_pertdata]
        train_groups, test_groups = np.split(
            np.unique(groups), [int(len(np.unique(groups)) * 0.8)]
        )
        # get the first 1_000 highly variable genes in the pert data which are in the
        # celltype-spec GRN
        hvg_ = dataset_genes_f[np.argsort(np.var(dataset_genes_celltype, 0))[::-1]]
        to_train_on = hvg_[
            np.where(np.isin(hvg_, gene_lists_celltype_filt[celltype]))[0][
                : args.n_genes_to_train_on
            ]
        ]

        score = []
        for i in tqdm(
            range((0 if args.test_mode == "val" else 1), args.n_genes_to_train_on, 2),
            total=args.n_genes_to_train_on // 2,
        ):
            selected_gene = to_train_on[i]

            index_of_gene = np.where(
                gene_lists_celltype_filt[celltype] == selected_gene
            )[0]

            top_sim_genes_ranked = top_k_celltype_filt[celltype][index_of_gene[0]]

            tfs_for_that_gene = np.array(
                gene_2_tf_dict[selected_gene.decode("utf-8")]
            ).astype(bytes)

            top_sim_tfs_ranked = top_sim_genes_ranked[
                np.isin(top_sim_genes_ranked, tfs_for_that_gene)
            ]

            top_sim_tfs_in_pert = top_sim_tfs_ranked[
                np.isin(top_sim_tfs_ranked, dataset_tf_f)
            ][: args.n_features]

            if len(top_sim_tfs_in_pert) == 0:
                score.append(0)
            else:
                top_sim_tfs_in_pert = top_sim_tfs_in_pert
                Y = dataset_genes_celltype[
                    :, np.where(dataset_genes_f == selected_gene)[0]
                ]
                X = dataset_tf_celltype[:, np.isin(dataset_tf_f, top_sim_tfs_in_pert)]

                X_train = X[np.isin(groups, train_groups)]
                X_test = X[np.isin(groups, test_groups)]
                Y_train = Y[np.isin(groups, train_groups)]
                Y_test = Y[np.isin(groups, test_groups)]

                linreg = LinearRegression().fit(X_train, Y_train)
                score_ = linreg.score(X_test, Y_test)
                if score_ < 0:
                    score.append(0)
                else:
                    score.append(score_)
        score_per_celltype.append(np.mean(score))

    print(args.path_to_embeddings)
    print(np.mean(score_per_celltype))
    return None


if __name__ == "__main__":
    main()



import h5torch
import numpy as np
import torch
from torch.nn.utils.rnn import pad_sequence
from lightning import LightningDataModule
from importlib.resources import files
from torch.distributions.binomial import Binomial
from torch.distributions.poisson import Poisson
from torch.distributions.normal import Normal
from einops import rearrange
from h5torch.dataset import sample_csr
from copy import deepcopy
from torch.utils.data import *
import math


class BentoDataModule(LightningDataModule):
    def __init__(self, config):
        super().__init__()
        self.config = deepcopy(config)

    def setup(self, stage):

        preprocessing_mapper = {
            "RankCounts": RankCounts,
            "CountsPerX": CountsPerX,
            "FixedNorm": FixedNorm,
            "LogP1": LogP1,
            "Bin": Bin,
            "DuplicateCells": DuplicateCells,
            "FilterTopGenes": FilterTopGenes,
            "FilterRandomGenes": FilterRandomGenes,
            "PoissonResample": PoissonResample,
            "GaussianResample": GaussianResample,
            "BinomialSubsample": BinomialSubsample,
            "Mask": Mask,
            "MolecularCV": MolecularCV,
            "CountsAsPositions": CountsAsPositions,
            "Copy": Copy,
            "FilterHVG": FilterHVG,
            "EliminateZeros": EliminateZeros,
        }

        if "input_processing" in self.config:
            processor = []
            for prepr in self.config["input_processing"]:
                prepr_type = prepr["type"]
                prepr_kwargs = {k: v for k, v in prepr.items() if k != "type"}
                processor.append(preprocessing_mapper[prepr_type](**prepr_kwargs))
        else:
            processor = []
        processor = SequentialPreprocessor(*processor)

        if not self.config.perturb_mode:
            processing_class = CellSampleProcessor
        else:
            processing_class = PerturbationCellSampleProcessor

        f = h5torch.File(self.config["data_path"])

        if self.config["in_memory"]:
            f = f.to_dict()

        if ("filter_highseqdepth" in self.config) and (
            self.config["filter_highseqdepth"] == True
        ):
            path = files("bento_sc.utils.data").joinpath("cxg_filter.npz")
            train_indices = np.load(path)["train"]
            val_indices = np.load(path)["val"]
            test_indices = np.load(path)["test"]
        else:
            train_indices = None
            val_indices = None
            test_indices = None

        self.train = h5torch.Dataset(
            f,
            sample_processor=processing_class(
                processor,
                return_zeros=self.config["return_zeros"],
                deterministic=(
                    False
                    if "deterministic" not in self.config
                    else self.config["deterministic"]
                ),
            ),
            subset=(
                train_indices if train_indices is not None else ("0/split", "train")
            ),
        )

        self.val = h5torch.Dataset(
            f,
            sample_processor=processing_class(
                processor,
                return_zeros=self.config["return_zeros"],
                deterministic=(
                    False
                    if "deterministic" not in self.config
                    else self.config["deterministic"]
                ),
            ),
            subset=(
                val_indices
                if val_indices is not None
                else ("0/split", ("val_sub" if self.config.val_sub else "val"))
            ),
        )

        self.test = h5torch.Dataset(
            f,
            sample_processor=processing_class(
                processor,
                return_zeros=self.config["return_zeros"],
                deterministic=(
                    False
                    if "deterministic" not in self.config
                    else self.config["deterministic"]
                ),
            ),
            subset=(test_indices if test_indices is not None else ("0/split", "test")),
        )

    def train_dataloader(self):
        batch_sampler = self.configure_batch_sampler(self.train, n_partitions=100)
        if batch_sampler is not None:
            extra_kwargs = {}
        else:
            extra_kwargs = {"batch_size": self.config.batch_size, "shuffle": True}

        return torch.utils.data.DataLoader(
            self.train,
            num_workers=self.config.n_workers,
            prefetch_factor=(
                None
                if "prefetch_factor" not in self.config
                else self.config.prefetch_factor
            ),
            pin_memory=True,
            collate_fn=BatchCollater(self.config.allow_padding),
            batch_sampler=batch_sampler,
            **extra_kwargs,
        )

    def val_dataloader(self):
        batch_sampler = self.configure_batch_sampler(self.val, n_partitions=1)
        if batch_sampler is not None:
            extra_kwargs = {}
        else:
            extra_kwargs = {"batch_size": self.config.batch_size, "shuffle": False}

        return torch.utils.data.DataLoader(
            self.val,
            num_workers=self.config.n_workers,
            pin_memory=True,
            prefetch_factor=(
                None
                if "prefetch_factor" not in self.config
                else self.config.prefetch_factor
            ),
            collate_fn=BatchCollater(self.config.allow_padding),
            batch_sampler=batch_sampler,
            **extra_kwargs,
        )

    def test_dataloader(self):
        batch_sampler = None

        extra_kwargs = {"batch_size": self.config.batch_size, "shuffle": False}

        return torch.utils.data.DataLoader(
            self.test,
            num_workers=self.config.n_workers,
            pin_memory=True,
            prefetch_factor=(
                None
                if "prefetch_factor" not in self.config
                else self.config.prefetch_factor
            ),
            collate_fn=BatchCollater(True),
            batch_sampler=batch_sampler,
            **extra_kwargs,
        )

    def configure_batch_sampler(self, dataset, n_partitions=100):
        if not self.config.return_zeros:
            if (
                isinstance(self.config.devices, list) and len(self.config.devices) > 1
            ) or (isinstance(self.config.devices, int) and self.config.devices > 1):
                batch_sampler = DistributedBucketSampler(
                    dataset, self.config.batch_size, n_partitions=n_partitions
                )
            else:
                batch_sampler = BucketBatchSampler(
                    dataset, self.config.batch_size, n_partitions=n_partitions
                )

        else:
            batch_sampler = None
        return batch_sampler

    def predict_dataloader(self):
        return self.test_dataloader()

    @property
    def config_used(self):
        return {
            "input_processing",
            "perturb_mode",
            "data_path",
            "return_zeros",
            "in_memory",
            "n_workers",
            "batch_size" "devices",
            "allow_padding",
        }

    @property
    def config_unused(self):
        return set(self.config) - self.config_used


class BatchCollater:
    def __init__(self, allow_padding=False):
        self.allow_padding = allow_padding

    def __call__(self, batch):
        batch_collated = {}

        batch_collated["0/split"] = [b["0/split"] for b in batch]

        if "0/obs" in batch[0]:
            batch_collated["0/obs"] = torch.tensor(
                np.array([b["0/obs"] for b in batch])
            )

        if "0/perturbed_gene" in batch[0]:
            batch_collated["0/perturbed_gene"] = torch.tensor(
                np.array([b["0/perturbed_gene"] for b in batch])
            ).long()

        if "0/ADT" in batch[0]:
            batch_collated["0/targets"] = torch.tensor(
                np.array([b["0/ADT"] for b in batch])
            )
        elif ("0/obs" in batch[0]) and (batch[0]["0/obs"].shape[0] == 9):
            # essentially hardcoded to detect the cellxgene pre-training set this way
            # (length 9 obs TODO make more elegant by changing the cellxgene processing script)
            batch_collated["0/targets"] = torch.tensor(
                np.array([b["0/obs"][3] for b in batch])
            )

        if "0/celltype" in batch[0]:
            batch_collated["0/celltype"] = [b["0/celltype"] for b in batch]

        counts_keys = ["gene_index", "gene_counts", "gene_counts_true"]
        append_value = [0, -1, -1]
        if "gene_counts_copy" in batch[0]:
            counts_keys.append("gene_counts_copy")
            append_value.append(-1)

        for name, padval in zip(counts_keys, append_value):
            samples_key = [b[name] for b in batch]
            if not self.allow_padding:
                samples_key = self.cut_to_uniform_size(samples_key)

            if batch[0][name].ndim == 1:
                if len({b[name].shape for b in batch}) == 1:
                    batch_collated[name] = torch.stack(samples_key)
                else:
                    batch_collated[name] = pad_sequence(
                        samples_key, batch_first=True, padding_value=padval
                    )
            else:
                if len({b[name].shape for b in batch}) == 1:
                    batch_collated[name] = torch.cat(samples_key, 0)
                else:
                    batch_collated[name] = rearrange(
                        pad_sequence(
                            [b.T for b in samples_key],
                            batch_first=True,
                            padding_value=padval,
                        ),
                        "b l k -> (b k) l",
                    )

        if len(batch_collated["gene_index"]) != len(batch_collated["0/split"]):
            batch_collated["0/split"] = list(np.repeat(batch_collated["0/split"], 2))
            if "0/obs" in batch_collated:
                batch_collated["0/obs"] = torch.repeat_interleave(
                    batch_collated["0/obs"], 2, dim=0
                )

        return batch_collated

    @staticmethod
    def cut_to_uniform_size(list_of_objects):
        min_len = min([b.shape[-1] for b in list_of_objects])
        return [b[..., :min_len] for b in list_of_objects]


class BucketBatchSampler(BatchSampler):
    def __init__(
        self,
        dataset,
        batch_size,
        n_partitions=100,
        indices=None,
        drop_last=False,
    ):
        super().__init__(dataset, batch_size, drop_last)

        self.len_dataset = len(dataset) if indices is None else len(indices)

        self.seqlens = torch.tensor(
            np.diff(dataset.f["central/indptr"][:])[dataset.indices]
        )
        if indices is not None:
            self.seqlens = self.seqlens[indices]
            indices = np.array(indices)
        else:
            indices = np.arange(self.len_dataset)

        self.bucket_sampler = BatchSampler(
            RandomSampler(indices), math.ceil(self.len_dataset / n_partitions), False
        )
        self.n_partitions = n_partitions
        self.indices = indices

    def __iter__(self):
        bucket_indices_all = []
        for bucket in self.bucket_sampler:
            bucket_indices = self.indices[bucket]
            bucket_asort_seqlens = torch.argsort(self.seqlens[bucket], descending=True)
            bucket_indices_in_order = list(bucket_indices[bucket_asort_seqlens.numpy()])
            bucket_indices_all.append(bucket_indices_in_order)

        for bucket_indices in bucket_indices_all:
            for batch in SubsetRandomSampler(
                list(BatchSampler(bucket_indices, self.batch_size, self.drop_last))
            ):  # SubsetRandomSampler(list())
                yield batch

    def __len__(self):
        t = [math.ceil(self.len_dataset / self.n_partitions)] * self.n_partitions
        t[-1] -= (self.len_dataset // self.n_partitions) % self.n_partitions
        len_ = sum([math.ceil(tt / self.batch_size) for tt in t])
        return len_


class DistributedBucketSampler(DistributedSampler):
    def __init__(
        self,
        dataset,
        batch_size,
        n_partitions=100,
        num_replicas=None,
        rank=None,
        shuffle=True,
        seed=0,
        drop_last=False,
    ):
        super().__init__(
            dataset,
            num_replicas=num_replicas,
            rank=rank,
            shuffle=shuffle,
            seed=seed,
            drop_last=drop_last,
        )

        self.batch_size = batch_size
        self.n_partitions = n_partitions

    def __iter__(self):
        indices = list(super().__iter__())
        batch_sampler = BucketBatchSampler(
            self.dataset,
            batch_size=self.batch_size,
            n_partitions=self.n_partitions,
            indices=indices,
        )
        return iter(batch_sampler)

    def __len__(self):
        t = [math.ceil(self.num_samples / self.n_partitions)] * self.n_partitions
        t[-1] -= (self.num_samples // self.n_partitions) % self.n_partitions
        len_ = sum([math.ceil(tt / self.batch_size) for tt in t])
        return len_


class CellSampleProcessor:
    def __init__(
        self,
        processor,
        return_zeros=False,
        n_genes=19331,
        deterministic=False,
    ):
        self.processor = processor

        self.return_zeros = return_zeros
        self.n_genes = n_genes
        self.deterministic = deterministic

    def __call__(self, f, sample):

        if self.deterministic:
            np.random.seed(42)

        if self.return_zeros:
            gene_counts = np.zeros(self.n_genes)
            gene_counts[sample["central"][0]] = sample["central"][1]
            sample |= {
                "gene_counts": torch.tensor(gene_counts),
                "gene_counts_true": torch.tensor(gene_counts),
                "gene_index": torch.arange(self.n_genes),
            }
        else:
            asort = np.argsort(
                sample["central"][1] + np.random.rand(len(sample["central"][1]))
            )[::-1]
            gene_counts = sample["central"][1][asort]
            gene_index = sample["central"][0][asort].view(np.ndarray).astype(np.int64)
            sample |= {
                "gene_counts": torch.tensor(gene_counts),
                "gene_index": torch.tensor(gene_index),
                "gene_counts_true": torch.tensor(gene_counts),
            }

        _ = sample.pop("central")
        if self.processor is not None:
            sample = self.processor(sample)
        return sample


class PerturbationCellSampleProcessor:
    def __init__(self, processor, return_zeros=True, n_genes=19331, deterministic=None):
        assert (
            return_zeros == True
        ), "return zeros has to be true for perturbation modeling."

        self.processor = processor
        self.processor_trues = SequentialPreprocessor(
            CountsPerX(factor=10_000, key="gene_counts_true"),
            LogP1(key="gene_counts_true"),
        )

        self.processor_origs = SequentialPreprocessor(
            Copy(key="gene_counts", to="gene_counts_copy"),
            CountsPerX(factor=10_000, key="gene_counts_copy"),
            LogP1(key="gene_counts_copy"),
        )

        self.n_genes = n_genes
        path = files("bento_sc.utils.data").joinpath("gene_set_perturb.txt")
        self.gene_indices = torch.tensor(np.loadtxt(path).astype(int))

    def __call__(self, f, sample):
        gene_counts = np.zeros(self.n_genes)
        gene_counts[sample["central"][0]] = sample["central"][1]
        sample |= {
            "gene_counts_true": torch.tensor(gene_counts),
            "gene_index": torch.arange(self.n_genes),
        }

        if sample["0/split"] != "train":
            control_sample = sample_csr(f["central"], sample["0/matched_control"])

        else:
            train_control_indices = f["unstructured/train_control_indices"][:]
            sampled_control = np.random.choice(train_control_indices)
            control_sample = sample_csr(f["central"], sampled_control)

        gene_counts = np.zeros(self.n_genes)
        gene_counts[control_sample[0]] = control_sample[1]
        sample["gene_counts"] = torch.tensor(gene_counts)

        _ = sample.pop("central")

        sample = self.processor_origs(sample)

        if self.processor is not None:
            sample = self.processor(sample)

        sample = self.processor_trues(sample)

        sample["gene_counts_copy"] = sample["gene_counts_copy"][self.gene_indices]
        sample["gene_counts"] = sample["gene_counts"][self.gene_indices]
        sample["gene_counts_true"] = sample["gene_counts_true"][self.gene_indices]
        sample["gene_index"] = self.gene_indices

        return sample


class RankCounts:
    def __init__(self, key="gene_counts"):
        self.key = key

    def __call__(self, sample):
        sample[self.key] = torch.argsort(sample[self.key], descending=True)
        return sample


class CountsPerX:
    def __init__(self, factor=10_000, key="gene_counts"):
        self.factor = factor
        self.key = key

    def __call__(self, sample):
        if sample[self.key].ndim == 2:
            total = sample[self.key].sum(1, keepdim=True)
        else:
            total = sample[self.key].sum()

        sample[self.key] = sample[self.key] / total * self.factor
        return sample


class FixedNorm:
    def __init__(self, factor=0.01, key="gene_counts"):
        self.factor = factor
        self.key = key

    def __call__(self, sample):
        sample[self.key] = sample[self.key] * self.factor
        return sample


class LogP1:
    def __init__(self, key="gene_counts"):
        self.key = key

    def __call__(self, sample):
        sample[self.key] = torch.log1p(sample[self.key])
        return sample


class Bin:
    def __init__(self, key="gene_counts"):
        bins = np.loadtxt(files("bento_sc.utils.data").joinpath("bins.txt"))
        self.bins = torch.tensor(bins)
        self.key = key

    def __call__(self, sample):
        sample[self.key] = torch.bucketize(sample[self.key], self.bins, right=True) - 1
        return sample

    @staticmethod
    def generate_bins(x, n):
        bins = np.quantile(x, np.linspace(0, 1, n + 1))
        bins = np.unique(bins)
        bins[-1] = bins[-1] + 1
        return bins


class DuplicateCells:
    def __init__(self, affected_keys=["gene_counts", "gene_index", "gene_counts_true"]):

        self.affected_keys = affected_keys

    def __call__(self, sample):
        shuffler = torch.randperm(len(sample[self.affected_keys[0]]))
        for a in self.affected_keys:
            sample[a] = torch.stack([sample[a], sample[a].clone()[shuffler]])
        return sample


class FilterTopGenes:
    def __init__(
        self,
        number=1024,
        affected_keys=["gene_counts", "gene_index", "gene_counts_true"],
        key_to_determine_top="gene_counts",
    ):
        self.n = number
        self.affected_keys = affected_keys
        self.topkey = key_to_determine_top

    def __call__(self, sample):
        if sample[self.affected_keys[0]].ndim == 1:
            to_select = torch.argsort(sample[self.topkey]).flip(0)[: self.n]
            for a in self.affected_keys:
                sample[a] = sample[a][to_select]
        else:
            to_select = torch.argsort(sample[self.topkey]).fliplr()[:, : self.n]
            for a in self.affected_keys:
                sample[a] = sample[a][torch.arange(2).unsqueeze(-1), to_select]
        return sample


class FilterHVG:
    def __init__(
        self,
        affected_keys=["gene_counts", "gene_index", "gene_counts_true"],
        number=1024,
        dataset="cellxgene",
    ):
        assert dataset in [
            "cellxgene",
            "citeseq",
            "greatapes",
            "embryoniclimb",
            "circimm",
        ]
        path = files("bento_sc.utils.data").joinpath("hvg_%s.npy" % dataset)
        var = np.load(path)
        asort = np.argsort(var)[::-1][:number]
        self.to_select = torch.sort(torch.tensor(asort.copy())).values
        self.affected_keys = affected_keys

    def __call__(self, sample):
        if sample[self.affected_keys[0]].ndim == 1:
            len_ = len(sample[self.affected_keys[0]])
            assert len_ == 19331
            indices = torch.argsort(sample["gene_index"])[self.to_select]
            for a in self.affected_keys:
                sample[a] = sample[a][indices]
        else:
            len_ = len(sample[self.affected_keys[0]][0])
            assert len_ == 19331
            indices = torch.argsort(sample["gene_index"])[:, self.to_select]
            for a in self.affected_keys:
                sample[a] = sample[a][torch.arange(2).unsqueeze(-1), indices]
        return sample


class FilterRandomGenes:
    def __init__(
        self,
        number=1024,
        affected_keys=["gene_counts", "gene_index", "gene_counts_true"],
        proportional_hvg=False,
    ):
        self.n = number
        self.affected_keys = affected_keys
        if proportional_hvg:
            path = files("bento_sc.utils.data").joinpath("hvg_cellxgene.npy")
            var = np.load(path)
            self.p = (
                np.exp(np.log10(var + 1e-8) / 5)
                / np.exp(np.log10(var + 1e-8) / 5).sum()
            )
        self.proportional_hvg = proportional_hvg

    def __call__(self, sample):
        if sample[self.affected_keys[0]].ndim == 1:
            len_ = len(sample[self.affected_keys[0]])
            indices = self._sample(sample, len_)
            for a in self.affected_keys:
                sample[a] = sample[a][indices]
        else:
            len_ = len(sample[self.affected_keys[0]][0])
            indices = torch.stack(
                [
                    self._sample(sample, len_),
                    self._sample(sample, len_),
                ]
            )
            for a in self.affected_keys:
                sample[a] = sample[a][torch.arange(2).unsqueeze(-1), indices]
        return sample

    def _sample(self, sample, len_):
        if not self.proportional_hvg:
            return torch.randperm(len_)[: self.n]
        else:
            assert (
                len_ == 19331
            ), "proportional hvg sampling can only be done on all genes"
            to_sample_gene_ix = torch.tensor(
                np.random.choice(len_, size=(self.n,), p=self.p), replace=False
            )
            return torch.argsort(sample["gene_index"])[to_sample_gene_ix]


class PoissonResample:
    def __init__(self, key="gene_counts", clamp_1=True):
        self.key = key
        self.clamp = clamp_1

    def __call__(self, sample):
        resampled = Poisson(sample[self.key]).sample()
        if self.clamp:
            resampled = torch.clamp(resampled, min=1)
        sample[self.key] = resampled
        return sample


class Copy:
    def __init__(self, key="gene_counts", to="gene_counts_copy"):
        self.key = key
        self.to = to

    def __call__(self, sample):
        sample[self.to] = sample[self.key].clone()
        return sample


class CountsAsPositions:
    def __init__(
        self,
        skip_trues=False,
    ):
        """
        Used for Geneformer-style pre-training:
            - Gene index: RankCounts Gene counts
            - Gene counts true: Gene index
            - Gene counts: Gene index
        To train Geneformer-style, make sure these have been run before performing this step:
            RankCounts gene counts
            FilterTop
        """
        self.skip_trues = skip_trues

    def __call__(self, sample):
        gene_index = sample["gene_index"].clone()
        gene_counts = sample["gene_counts"].clone()
        sample["gene_index"] = gene_counts
        sample["gene_counts"] = gene_index
        if not self.skip_trues:
            sample["gene_counts_true"] = gene_index
        return sample


class EliminateZeros:
    def __init__(
        self,
        key="gene_counts",
        affected_keys=["gene_counts", "gene_index", "gene_counts_true"],
    ):
        """
        Used to make MCV compatible with non-zero transformer input.
        """
        self.key = key
        self.affected_keys = affected_keys

    def __call__(self, sample):
        assert sample[self.key].ndim == 1

        keep = sample[self.key] != 0

        for a in self.affected_keys:
            sample[a] = sample[a][keep]

        return sample


class GaussianResample:
    def __init__(self, key="gene_counts", std=1):
        self.key = key
        self.std = std

    def __call__(self, sample):
        sample[self.key] = torch.maximum(
            Normal(sample[self.key], self.std).sample(), sample[self.key].min()
        )
        return sample


class BinomialSubsample:
    def __init__(self, key="gene_counts", p=0.75, clamp_1=True):
        self.clamp = clamp_1
        self.key = key
        self.p = torch.tensor(p)

    def __call__(self, sample):
        resampled = Binomial(sample[self.key], self.p).sample()
        if self.clamp:
            resampled = torch.clamp(resampled, min=1)
        sample[self.key] = resampled
        return sample


class Mask:
    def __init__(self, p=0.15, key="gene_counts"):
        self.p = p
        self.key = key

    def __call__(self, sample):
        replace = sample[self.key].float()
        to_mask = torch.rand_like(replace) < self.p
        replace[to_mask] = torch.nan
        sample[self.key] = replace
        return sample


class MolecularCV:
    def __init__(self, p_to_train=0.1):
        self.p = torch.tensor(p_to_train)

    def __call__(self, sample):
        resampled = Binomial(sample["gene_counts"], self.p).sample()
        sample["gene_counts_true"] = sample["gene_counts"] - resampled
        sample["gene_counts"] = resampled
        return sample


class SequentialPreprocessor:
    def __init__(self, *args):
        self.preprocessors = args

    def __call__(self, sample):
        for step in self.preprocessors:
            sample = step(sample)
        return sample



import os

os.environ["PYTORCH_CUDA_ALLOC_CONF"] = "max_split_size_mb:2048"

from bento_sc.data import BentoDataModule
from bento_sc.models import BentoTransformer
from bento_sc.utils.config import Config
from lightning.pytorch.callbacks import ModelCheckpoint
from lightning.pytorch.loggers import TensorBoardLogger
from lightning.pytorch.plugins.environments import LightningEnvironment
from lightning.pytorch import Trainer
import argparse


def boolean(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


def main():
    class CustomFormatter(
        argparse.ArgumentDefaultsHelpFormatter, argparse.MetavarTypeHelpFormatter
    ):
        pass

    parser = argparse.ArgumentParser(
        description="Pre-training script.",
        formatter_class=CustomFormatter,
    )

    parser.add_argument(
        "config_path", type=str, metavar="config_path", help="config_path"
    )
    parser.add_argument("logs_path", type=str, metavar="logs_path", help="logs_path")
    parser.add_argument(
        "--data_path",
        type=str,
        default=None,
        help="Data file. Overrides value in config file if specified",
    )
    parser.add_argument(
        "--lr",
        type=float,
        default=None,
        help="Learning rate. Overrides value in config file if specified",
    )
    parser.add_argument(
        "--ckpt_path", type=str, default=None, help="Continue from checkpoint"
    )
    parser.add_argument(
        "--tune_mode",
        type=boolean,
        default=False,
        help="Don't pre-train whole model but run small experiment.",
    )

    args = parser.parse_args()

    config = Config(args.config_path)

    if args.data_path is not None:
        config["data_path"] = args.data_path
    if args.lr is not None:
        config["lr"] = args.lr

    dm = BentoDataModule(config)
    dm.setup(None)

    model = BentoTransformer(config)

    callbacks = [
        ModelCheckpoint(every_n_train_steps=5000),
    ]
    logger = TensorBoardLogger(
        "/".join(args.logs_path.split("/")[:-1]),
        name=args.logs_path.split("/")[-1],
    )

    if args.tune_mode:
        max_steps = 2_501
        val_check_interval = 250
    else:
        max_steps = 200_000
        val_check_interval = 5_000
    if ("no_genewise_loss" in config) and (config["no_genewise_loss"] == True):
        strategy = "ddp_find_unused_parameters_true"
    else:
        strategy = "auto"
    trainer = Trainer(
        accelerator="gpu",
        devices=config.devices,
        strategy=strategy,
        plugins=[LightningEnvironment()],
        gradient_clip_val=1,
        max_steps=max_steps,
        val_check_interval=val_check_interval,
        check_val_every_n_epoch=None,
        callbacks=callbacks,
        logger=logger,
        precision="bf16-true",
        use_distributed_sampler=(True if config.return_zeros else False),
    )

    trainer.fit(model, dm, ckpt_path=args.ckpt_path)


if __name__ == "__main__":
    main()
